@@ -1,4 +1,7 @@
-/* $Id: acconfig.h,v 1.2 1998/03/02 08:33:14 tudor Exp $ */
+/* $Id: acconfig.h,v 1.3 1999/01/16 22:37:15 tudor Exp $ */
+
+/* Define if you want to disable various consistency checkings.  */
+#undef NDEBUG
 
 /* Define if you have GCC.  */
 #undef HAVE_GCC
@@ -50,6 +53,15 @@
 /* Define if you have the statvfs system call.  */
 #undef STAT_STATVFS
 
+/* Define if you have a broken sys/filesys.h header file.  */
+#undef BROKEN_SYS_FILSYS_H
+
+/* Define this to the name of the package.  */
+#undef PACKAGE
+
+/* Define this to the version of the package.  */
+#undef VERSION
+
 
 /* Leave that blank line there!!  Autoheader needs it.
    If you're adding to this file, keep in mind:
