@@ -23,7 +23,7 @@
 
 /* This tells Alpha OSF/1 not to define a getopt prototype in <stdio.h>.
    Ditto for AIX 3.2 and <stdlib.h>.  */
-/* $Id: getopt.c,v 1.3 1998/03/10 05:12:11 tudor Exp $ */
+/* $Id: getopt.c,v 1.4 1999/05/31 18:51:22 tudor Exp $ */
 #ifndef _NO_PROTO
 #define _NO_PROTO
 #endif
@@ -449,6 +449,10 @@ _getopt_initialize (argc, argv, optstring)
     }
   else
     nonoption_flags_len = 0;
+#else
+  /* Avoid `unused variable' warnings.  */  
+  argc = 0;
+  argv = NULL;
 #endif
 
   return optstring;
@@ -697,6 +701,7 @@ _getopt_internal (argc, argv, optstring, longopts, longind, long_only)
 	      else
 		{
 		  if (opterr)
+		  {
 		   if (argv[optind - 1][1] == '-')
 		    /* --option */
 		    fprintf (stderr,
@@ -707,6 +712,7 @@ _getopt_internal (argc, argv, optstring, longopts, longind, long_only)
 		    fprintf (stderr,
 		     _("%s: option `%c%s' doesn't allow an argument\n"),
 		     argv[0], argv[optind - 1][0], pfound->name);
+		  }
 
 		  nextchar += strlen (nextchar);
 
