@@ -1,6 +1,6 @@
 /* git.c -- The main git file.  */
 
-/* Copyright (C) 1993-1998 Free Software Foundation, Inc.
+/* Copyright (C) 1993-1999 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -17,7 +17,7 @@
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
 /* Written by Tudor Hulubei and Andrei Pitis.  */
-/* $Id: git.c,v 1.7 1998/03/10 05:12:11 tudor Exp $ */
+/* $Id: git.c,v 1.19 1999/05/31 21:21:03 tudor Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include <config.h>
@@ -73,7 +73,7 @@ char *copyright =
 "GIT is free software; you can redistribute it and/or modify it under the\n\
 terms of the GNU General Public License as published by the Free Software\n\
 Foundation; either version 2, or (at your option) any later version.\n\
-Copyright (C) 1993-1998 Free Software Foundation, Inc.\n\
+Copyright (C) 1993-1999 Free Software Foundation, Inc.\n\
 Written by Tudor Hulubei and Andrei Pitis, Bucharest, Romania\n\n";
 
 
@@ -102,7 +102,7 @@ int TypeSensitivity = ON;
 
 char *home;
 char *program;
-char *version=VERSION;
+char *version = VERSION;
 int two_panel_mode = 1;
 int current_mode = GIT_SCREEN_MODE;
 int panel_no;
@@ -113,9 +113,7 @@ int UseLastScreenChar;
 char color_section[]  = "[GIT-Color]";
 char monochrome_section[] = "[GIT-Monochrome]";
 
-char lock_think[] = "Wait, I am thinking...";
-char lock_ok[]    = "Wait, I am thinking... well, it looks good.";
-char lock_bad[]   = "Wait, I am thinking... well, lets try again.";
+char lock_bad[]   = "Bad password, try again...";
 
 char *exit_msg;
 
@@ -132,7 +130,7 @@ char **dir_history;
 int dir_history_count;
 int dir_history_point;
 
-#define BUILTIN_OPERATIONS                       85
+#define BUILTIN_OPERATIONS                       87
 
 #define BUILTIN_copy				 -1
 #define BUILTIN_move				 -2
@@ -219,7 +217,8 @@ int dir_history_point;
 #define BUILTIN_bin_packing			-83
 #define BUILTIN_horizontal_scroll_left		-84
 #define BUILTIN_horizontal_scroll_right		-85
-
+#define BUILTIN_select_extension		-86
+#define BUILTIN_unselect_extension		-87
 
 #define MAX_BUILTIN_NAME			 35
 
@@ -311,6 +310,8 @@ char builtin[BUILTIN_OPERATIONS][MAX_BUILTIN_NAME] =
     "bin-packing",
     "horizontal-scroll-left",
     "horizontal-scroll-right",
+    "select-extension",
+    "unselect-extension",
 };
 
 
@@ -338,7 +339,6 @@ typedef struct
 /*
  * Return ON if there is enough space on the screen to display the panels.
  */
-
 int
 panels_can_be_displayed()
 {
@@ -358,10 +358,19 @@ panels_can_be_displayed()
 }
 
 
+/*
+ * Return 1 if in GIT_TERMINAL_MODE.
+ */
+int
+in_terminal_mode()
+{
+    return (current_mode == GIT_TERMINAL_MODE);
+}
+
+
 /*
  * Resize (if necessary) all git's components.
  */
-
 void
 resize(resize_required)
     int resize_required;
@@ -449,7 +458,6 @@ resize(resize_required)
  * (SIGCONT or SIGWINCH) but can be 0 if unconditional refresh is
  * desired.
  */
-
 void
 refresh(signum)
     int signum;
@@ -492,7 +500,8 @@ refresh(signum)
 	tty_put_screen(screen);
 
     status_update();
-    il_full_update();
+    il_update();
+    il_update_point();
     tty_update();
 
     if (signum == SIGCONT)
@@ -500,6 +509,37 @@ refresh(signum)
 }
 
 
+void
+report_undefined_key(status_message)
+    char *status_message;
+{
+    char *prev = tty_get_previous_key_seq();
+    size_t length = strlen(prev);
+
+    if (length && (prev[length - 1] != key_INTERRUPT))
+    {
+	char *str = (char *)tty_key_machine2human(prev);
+	char *buf = xmalloc(128 + strlen(str));
+	sprintf(buf, "%s: not defined.", str);
+	status(buf, STATUS_ERROR, STATUS_LEFT);
+	xfree(buf);
+
+	tty_beep();
+	tty_update();
+	sleep(1);
+    }
+    else
+	tty_beep();
+
+    if (status_message)
+	status(status_message, STATUS_OK, STATUS_CENTERED);
+    else
+	status_default();
+    il_update_point();
+    tty_update();
+}
+
+
 /*****************************************/
 /* The GIT interface to the input line.  */
 /*****************************************/
@@ -542,7 +582,6 @@ il_history_add_entry(history, text)
 /*
  * Preview a history string.
  */
-
 char *
 il_history_view_entry(history, offset)
     xstack_t *history;
@@ -561,7 +600,6 @@ il_history_view_entry(history, offset)
  * of the input line.  If IL_MOVE is not specified, the  IL_EDIT  flag is
  * ignored.  Returns 1 if key has been processed and 0 otherwise.
  */
-
 int
 il_dispatch_commands(key, flags)
     int key;
@@ -695,7 +733,6 @@ il_dispatch_commands(key, flags)
  * and expand tabs.  Return a malloc-ed pointer to the fixed text.
  * The caller should free the new text.
  */
-
 char *
 il_fix_text(text)
    char *text;
@@ -718,7 +755,7 @@ il_fix_text(text)
 	    j += 8;
 	}
 	else
-	    if (isprint(text[i]))
+	    if (isprint((int)text[i]))
 		fixed_text[j++] = text[i];
 	    else
 		fixed_text[j++] = '?';
@@ -762,7 +799,6 @@ il_build_help_from_string(options)
  * answer.  The default char is the first char in the options string.
  * Returns 0 if it was interrupted, a valid character otherwise.
  */
-
 char
 il_read_char(message, options, flags)
     char *message;
@@ -805,16 +841,23 @@ il_read_char(message, options, flags)
 	}
     }
 
-    il_full_update();
+    il_update();
+    il_update_point();
+    tty_update();
 
     if (flags & IL_BEEP)
 	tty_beep();
 
     while (1)
     {
-	ks  = tty_get_key(&repeat_count);
-	key = ks->key_seq[0];
+	while ((ks = tty_get_key(&repeat_count)) == NULL)
+	{
+	    tty_beep();
+	    status_update();
+	    il_update_point();
+	}
 
+	key = ks->key_seq[0];
 	command = (command_t *)ks->aux_data;
 
 	if (command && command->builtin)
@@ -852,7 +895,9 @@ il_read_char(message, options, flags)
 		break;
 	}
 
+	status_update();
 	il_update_point();
+	tty_update();
     }
 
   done:
@@ -861,7 +906,9 @@ il_read_char(message, options, flags)
     if ((flags & IL_SAVE) && saved_il)
     {
 	il_restore(saved_il);
-	il_full_update();
+	il_update();
+	il_update_point();
+	tty_update();
     }
 
     return (key == key_INTERRUPT) ? 0 : key;
@@ -875,7 +922,6 @@ il_read_char(message, options, flags)
  * the second case, it will reallocate the pointer as needed using
  * xrealloc().  The caller should free the memory allocated this way.
  */
-
 char *
 il_read_line(static_text, dest, default_string, history)
     char *static_text;
@@ -886,7 +932,7 @@ il_read_line(static_text, dest, default_string, history)
     tty_key_t *ks;
     char *history_text;
     command_t *command;
-    int key, repeat_count, offset = 0;
+    int key = 0, repeat_count, offset = 0;
 
 
     il_reset_line();
@@ -903,13 +949,20 @@ il_read_line(static_text, dest, default_string, history)
 	offset = 1;
     }
 
-    il_full_update();
+    il_update();
+    il_update_point();
+    tty_update();
 
     while (1)
     {
-	ks  = tty_get_key(&repeat_count);
-	key = ks->key_seq[0];
+	while ((ks = tty_get_key(&repeat_count)) == NULL)
+	{
+	    tty_beep();
+	    status_update();
+	    il_update_point();
+	}
 
+	key = ks->key_seq[0];
 	command = (command_t *)ks->aux_data;
 
 	if (command && command->builtin)
@@ -933,7 +986,8 @@ il_read_line(static_text, dest, default_string, history)
 		{
 		    il_kill_line(IL_DONT_STORE);
 		    il_insert_text(history_text);
-		    il_full_update();
+		    il_update();
+		    il_update_point();
 		}
 		break;
 
@@ -945,7 +999,8 @@ il_read_line(static_text, dest, default_string, history)
 		if (offset == 0)
 		{
 		    il_kill_line(IL_DONT_STORE);
-		    il_full_update();
+		    il_update();
+		    il_update_point();
 		    break;
 		}
 
@@ -959,7 +1014,8 @@ il_read_line(static_text, dest, default_string, history)
 		    il_insert_text(history_text);
 		}
 
-		il_full_update();
+		il_update();
+		il_update_point();
 		break;
 
 	    case BUILTIN_refresh:
@@ -981,7 +1037,9 @@ il_read_line(static_text, dest, default_string, history)
 		break;
 	}
 
+	status_update();
 	il_update_point();
+	tty_update();
     }
 
   done:
@@ -1006,7 +1064,7 @@ il_read_line(static_text, dest, default_string, history)
  *					   so if there is no matching element
  *					   in the panel stack, we should delete
  *					   the last character in the input line
- * *action = IL_ISEARCH_ACTION_RETRY    -> the user pressed the isearch-forward
+ * *action = IL_ISEARCH_ACTION_RETRY    -> the user pressed the isearch-*
  *					   character again, so we should try to
  *					   find a new match for the current
  *					   string
@@ -1014,7 +1072,6 @@ il_read_line(static_text, dest, default_string, history)
  *					   into the input line so we should try
  *					   to find a match for the new string
  */
-
 char *
 il_isearch(static_text, dest, status, action)
     char *static_text;
@@ -1027,7 +1084,6 @@ il_isearch(static_text, dest, status, action)
     command_t *command;
     static input_line_t *saved_il;
 
-
     if (status == IL_ISEARCH_BEGIN)
     {
 	saved_il = il_save();
@@ -1042,7 +1098,9 @@ il_isearch(static_text, dest, status, action)
     if (status == IL_ISEARCH_END)
     {
 	il_restore(saved_il);
-	il_full_update();
+	il_update();
+	il_update_point();
+	tty_update();
 	return NULL;
     }
 
@@ -1051,10 +1109,19 @@ il_isearch(static_text, dest, status, action)
 
     *action = IL_ISEARCH_ACTION_NONE;
 
-    il_full_update();
+    il_update();
+    il_update_point();
+    tty_update();
 
   restart:
-    ks  = tty_get_key(NULL);
+    if ((ks = tty_get_key(NULL)) == NULL)
+    {
+	status_update();
+	il_update_point();
+	tty_update();
+	return NULL;
+    }
+
     key = ks->key_seq[0];
 
     command = (command_t *)ks->aux_data;
@@ -1094,11 +1161,11 @@ il_isearch(static_text, dest, status, action)
 		(key == BUILTIN_isearch_forward  &&
 		 status == IL_ISEARCH_FORWARD))
 	    {
-		if (!il_is_empty())
-		{
+		if (il_is_empty())
+		    *action = IL_ISEARCH_ACTION_INCREASE;
+		else
 		    *action = IL_ISEARCH_ACTION_RETRY;
-		    break;
-		}
+		break;
 	    }
 
 	    if (isprint(key))
@@ -1107,12 +1174,19 @@ il_isearch(static_text, dest, status, action)
 		*action = IL_ISEARCH_ACTION_INCREASE;
 	    }
 	    else
-		key = key_INTERRUPT;    /* Force a NULL return value.  */
+	    {
+		/* Force a NULL return value when detecting non printable
+		   characters (or functional keys).  */
+		key = key_INTERRUPT;
+	    }
 
 	    break;
     }
 
-    il_full_update();
+    status_update();
+    il_update();
+    il_update_point();
+    tty_update();
     il_get_contents(dest);
 
     return (key == BUILTIN_action || key == key_INTERRUPT) ? NULL : *dest;
@@ -1123,7 +1197,6 @@ il_isearch(static_text, dest, status, action)
 /* The directory history function set.  */
 /****************************************/
 
-
 void
 dir_history_reset()
 {
@@ -1212,13 +1285,12 @@ fatal(postmsg)
 
 /*
  * This function is a mess. Don't try to understand what it does
- * ... :-( It basically expands a configuration line macros.  The
+ * ... :-(  Basically, it expands a configuration line macro.  The
  * return value is 0 on error, -1 if some condition failed (the
  * command contains a %d but the current entry is not a directory), 1
  * if everything is ok, 2 if the command was correctly expanded and it
  * contains a '%i' and 3 if it contains a '%I'.  Easy, isn't it?
  */
-
 int
 command_expand(command, dest, p, l)
     command_t *command;
@@ -1251,12 +1323,11 @@ command_expand(command, dest, p, l)
 	    *d++ = *src++;
 	else
 	{
-	    t = islower(*++src) ? p : l;
+	    t = islower((int)*++src) ? p : l;
 
 	    switch (*src)
 	    {
 		case '?':
-
 		    if (busy)
 		    {
 			busy = 0;
@@ -1281,7 +1352,6 @@ command_expand(command, dest, p, l)
 		    break;
 
 		case 's':
-
 		    if (busy)
 		    {
 			busy = 0;
@@ -1365,12 +1435,10 @@ command_expand(command, dest, p, l)
 
 		case 'f':
 		case 'F':
-
 		    if (panel_get_current_file_type(t) != FILE_ENTRY)
 			goto strings_dont_match;
 
 		  get_file_name:
-
 		    ptr = panel_get_current_file_name(t);
 		    tmp = xmalloc(1 + strlen(ptr) + 1 + 1);
 		    sprintf(tmp, "\"%s\"", ptr);
@@ -1378,40 +1446,34 @@ command_expand(command, dest, p, l)
 
 		case 'd':
 		case 'D':
-
 		    if (panel_get_current_file_type(t) != DIR_ENTRY)
 			goto strings_dont_match;
 		    goto get_file_name;
 
 		case 'l':
 		case 'L':
-
 		    if (panel_get_current_file_type(t) != SYMLINK_ENTRY)
 			goto strings_dont_match;
 		    goto get_file_name;
 
 		case 't':
 		case 'T':
-
 		    if (panel_get_current_file_type(t) != FIFO_ENTRY)
 			goto strings_dont_match;
 		    goto get_file_name;
 
 		case 'z':
 		case 'Z':
-
 		    if (panel_get_current_file_type(t) != SOCKET_ENTRY)
 			goto strings_dont_match;
 		    goto get_file_name;
 
 		case 'a':
 		case 'A':
-
 		    goto get_file_name;
 
 		case 'm':
 		case 'M':
-
 		    tmp = xmalloc(16);
 		    sprintf(tmp, "%o",
 			    (int)panel_get_current_file_mode(t) & 07777);
@@ -1419,7 +1481,6 @@ command_expand(command, dest, p, l)
 
 		case 'o':
 		case 'O':
-
 		    uid = panel_get_current_file_uid(t);
 		    pwd = getpwuid(uid);
 
@@ -1435,7 +1496,6 @@ command_expand(command, dest, p, l)
 
 		case 'g':
 		case 'G':
-
 		    gid = panel_get_current_file_gid(t);
 		    grp = getgrgid(gid);
 
@@ -1451,14 +1511,12 @@ command_expand(command, dest, p, l)
 
 		case 'p':
 		case 'P':
-
 		    tmp = xmalloc(1 + strlen(t->path) + 1 + 1);
 		    sprintf(tmp, "\"%s\"", t->path);
 		    break;
 
 		case 'b':
 		case 'B':
-
 		    ptr = strrchr(t->path, '/');
 		    ptr = (*++ptr) ? ptr : "/root";
 		    tmp = xmalloc(1 + strlen(ptr) + 1 + 1);
@@ -1467,7 +1525,6 @@ command_expand(command, dest, p, l)
 
 		case 'i':
 		case 'I':
-
 		    i_flag = (*src == 'i') ? 1 : 2;
 
 		    if (busy && t->selected_entries)
@@ -1503,7 +1560,6 @@ command_expand(command, dest, p, l)
 		    break;
 
 		default:
-
 		    goto bad_command;
 	    }
 
@@ -1560,7 +1616,6 @@ add_to_environment(variable, alternate_variable, value)
  * Read keys from the current section ([GIT-Keys] is supposed to be in
  * use when read_keys() is called).  Return the number of keys read.
  */
-
 int
 read_keys(keys, errors)
     int keys;
@@ -1601,9 +1656,9 @@ read_keys(keys, errors)
 	    else
 	    {
 		/* This is not a TERMCAP symbol, it is a key sequence
-		   that we will have to convert it with
-		   tty_key_convert() into a machine usable form before
-		   using it.  */
+		   that we will have to be converted with
+		   tty_key_human2machine() into a machine usable form
+		   before being used.  */
 		need_conversion = 1;
 	    }
 	}
@@ -1631,22 +1686,22 @@ read_keys(keys, errors)
 	    goto insert;
 
 	if (contents[3])
-	    command->save_screen = ((tolower(contents[3][0]) == 'y') ? 1 : 0);
+	    command->save_screen = ((tolower((int)contents[3][0])=='y')?1:0);
 	else
 	    command->save_screen = 1;
 
 	if (contents[4])
-	    command->pause = ((tolower(contents[4][0]) == 'y') ? 1:0);
+	    command->pause = ((tolower((int)contents[4][0])=='y')?1:0);
 
 	if (contents[5])
-	    command->hide = ((tolower(contents[5][0]) == 'y') ? 1:0);
+	    command->hide = ((tolower((int)contents[5][0])=='y')?1:0);
 
       insert:
 	/* This speeds up the process.  It is not a limitation because,
 	   by convention, all the build-in command names contain only
 	   lowercase letters.  Avoid searching through the list of
 	   built-in command names.  */
-	if (islower(command->name[0]))
+	if (islower((int)command->name[0]))
 	    for (j = 0; j < BUILTIN_OPERATIONS; j++)
 	    {
 		if (strcmp(command->name, builtin[j]) == 0)
@@ -1664,7 +1719,7 @@ read_keys(keys, errors)
 	{
 	    if (need_conversion)
 	    {
-		if (tty_key_convert((unsigned char *)key_seq))
+		if (tty_key_human2machine((unsigned char *)key_seq))
 		    tty_key_list_insert((unsigned char *)key_seq,
 					(void *)command);
 		else
@@ -1686,7 +1741,6 @@ read_keys(keys, errors)
 /*
  * Last moment settings before suspending git.
  */
-
 void
 hide()
 {
@@ -1728,7 +1782,6 @@ usage()
     printf("usage: %s [-hvcblp] [path1] [path2]\n", program);
     printf(" -h         print this help message\n");
     printf(" -v         print the version number\n");
-    printf(" -i         print the installation directory\n");
     printf(" -c         use ANSI colors\n");
     printf(" -b         don't use ANSI colors\n");
     printf(" -l         don't use the last screen character\n");
@@ -1747,16 +1800,17 @@ main(argc, argv)
     size_t len = 0, ptrlen;
     char *temporary_directory;
     int previous_isearch_failed;
+    int resuming_previous_isearch;
     int output_final_path = OFF;
     input_line_t *saved_il = NULL;
     char *panel_path, *current_path;
     char *lock_password, *unlock_password;
     int child_exit_code, repeat_count, keys;
-    char *cmdln = NULL, *input = NULL, *ptr, *srcptr;
     int action_status, i, retval, to_case, cmp_mode, r1, r2;
     int c, ansi_colors = -1, use_last_screen_character = ON;
     int entry, key, app_end = 0, first_time = 1, errors = 0;
     char *left_panel_path, *right_panel_path, *output_string;
+    char *cmdln = NULL, *input = NULL, *ptr, *srcptr, *search_string = NULL;
 
 
     /* Make sure we don't get signals before we are ready to handle
@@ -1781,7 +1835,7 @@ main(argc, argv)
 	ansi_colors = ON;
 
     /* Parse the command line.  */
-    while ((c = getopt(argc, argv, "hvicblp")) != -1)
+    while ((c = getopt(argc, argv, "hvcblp")) != -1)
 	switch (c)
 	{
 	    case 'h':
@@ -1794,11 +1848,6 @@ main(argc, argv)
 		printf("%s %s\n", PRODUCT, VERSION);
 		return 0;
 
-	    case 'i':
-		/* Installation directory request.  */
-		printf("%s is installed in %s.\n", PRODUCT, PREFIX);
-		return 0;
-
 	    case 'c':
 		/* Force git to use ANSI color sequences.  */
 		ansi_colors = ON;
@@ -1857,11 +1906,12 @@ main(argc, argv)
 
     printf(copyright);
 
-    add_to_environment("GIT_EDITOR", "EDITOR",     "vi");
-    add_to_environment("GIT_SHELL",  "SHELL",      "/bin/sh");
-    add_to_environment("GIT_RMAIL",  (char *)NULL, "mail");
-    add_to_environment("GIT_PAGER",  (char *)NULL, "more");
-    add_to_environment("GIT_VMSTAT", (char *)NULL, "free");
+    add_to_environment("GIT_EDITOR",  "EDITOR",     "vi");
+    add_to_environment("GIT_SHELL",   "SHELL",      "/bin/sh");
+    add_to_environment("GIT_RMAIL",   (char *)NULL, "mail");
+    add_to_environment("GIT_PAGER",   (char *)NULL, "more");
+    add_to_environment("GIT_VMSTAT",  (char *)NULL, "free");
+    add_to_environment("GIT_BROWSER", (char *)NULL, "lynx");
 
     tty_init(TTY_RESTRICTED_INPUT);
 
@@ -2007,6 +2057,7 @@ main(argc, argv)
 	wait_msg = 0;
     }
 
+    tty_update_title(panel_get_path(src_panel));
     alarm(60 - get_local_time()->tm_sec);
 
     src_panel = panel_no ? right_panel : left_panel;
@@ -2021,8 +2072,9 @@ main(argc, argv)
     }
 
     title_update();
-    status_reset();
+    status_default();
     il_restore(saved_il);
+    tty_update();
 
     /* Save the input line contents.  */
     saved_il = il_save();
@@ -2048,13 +2100,18 @@ main(argc, argv)
     {
 	il_restore(saved_il);
 	saved_il = il_save();
-	il_full_update();
+	il_update();
+	il_update_point();
+	tty_update();
 	il_get_contents(&cmdln);
 
 	user_heart_attack = 0;
-	ks  = tty_get_key(&repeat_count);
-	key = ks->key_seq[0];
 
+	while ((ks = tty_get_key(&repeat_count)) == NULL)
+	    report_undefined_key(NULL);
+	status_update();
+
+	key = ks->key_seq[0];
 	command = (command_t *)ks->aux_data;
 
 	if (command)
@@ -2084,7 +2141,8 @@ main(argc, argv)
 				char *msg = xmalloc(msglen);
 
 				sprintf(msg, "%s: %s", command->name, cmd);
-				status(msg, MSG_STATUS, MSG_LEFT_ALIGNED);
+				status(msg, STATUS_WARNING, STATUS_LEFT);
+				tty_update();
 				xfree(msg);
 
 				if (command->hide)
@@ -2094,9 +2152,12 @@ main(argc, argv)
 					    "Wait, running %s command %s...",
 					    "user-defined", command->name);
 				    il_message(msg);
+				    tty_update();
 				    xfree(msg);
 				}
 
+				if (!is_a_bg_command(cmd))
+				    tty_update_title(cmd);
 				child_exit_code = start(cmd, command->hide);
 				xfree(cmd);
 
@@ -2140,7 +2201,6 @@ main(argc, argv)
 					    panel_unselect_all(dst_panel);
 				}
 
-				tty_update_title(panel_get_path(src_panel));
 				goto restart;
 			    }
 			    else
@@ -2171,7 +2231,6 @@ main(argc, argv)
 
 			    dir_history_add(panel_get_path(src_panel));
 			    xfree(expanded_dir);
-			    tty_update_title(panel_get_path(src_panel));
 			}
 
 			goto restart;
@@ -2179,7 +2238,7 @@ main(argc, argv)
 		}
 	    }
 	}
-	
+
 	switch (key)
 	{
 	    case key_INTERRUPT:
@@ -2254,12 +2313,6 @@ main(argc, argv)
 			il_kill_line(IL_DONT_STORE);
 			break;
 
-		    case '*':
-			panel_action(src_panel, act_TOGGLE,
-				     dst_panel, NULL, 1);
-			il_kill_line(IL_DONT_STORE);
-			break;
-
 		    case '\0':
 			action_status = panel_action(src_panel, act_ENTER,
 						     dst_panel, screen, 1);
@@ -2268,6 +2321,16 @@ main(argc, argv)
 			set_prompt();
 			break;
 
+		    case '*':
+			if (cmdln[1] == '\0')
+			{
+			    panel_action(src_panel, act_TOGGLE,
+					 dst_panel, NULL, 1);
+			    il_kill_line(IL_DONT_STORE);
+			    break;
+			}
+			/* Fall through... */
+
 		    default:
 			if (history_expand(cmdln, &output_string) >= 0)
 			{
@@ -2283,14 +2346,14 @@ main(argc, argv)
 			    }
 
 			    bg_cmd = is_a_bg_command(output_string);
-
+			    if (!bg_cmd)
+				tty_update_title(output_string);
 			    il_kill_line(IL_DONT_STORE);
 			    il_insert_text(output_string);
 			    start(output_string, bg_cmd);
 			    il_history(IL_RECORD);
 			    il_kill_line(IL_DONT_STORE);
-
-			    tty_update_title(panel_get_path(src_panel));
+			    /* HACK: Do not call tty_update(); here!  */
 
 			    if (!bg_cmd)
 			    {
@@ -2324,7 +2387,6 @@ main(argc, argv)
 	    case BUILTIN_scroll_down:
 		for (i = 0; i < repeat_count; i++)
 		    panel_action(src_panel, act_PGUP, dst_panel, NULL, 1);
-
 		break;
 
 	    case BUILTIN_scroll_up:
@@ -2352,20 +2414,26 @@ main(argc, argv)
 
 		alarm(0);
 		tty_put_screen(screen);
-		status(CommandLineModeHelp, MSG_OK, MSG_CENTERED);
+		status(CommandLineModeHelp, STATUS_OK, STATUS_CENTERED);
+		tty_update();
 
 		while (1)
 		{
 		    il_restore(saved_il);
 		    saved_il = il_save();
-		    il_full_update();
+		    il_update();
+		    il_update_point();
+		    tty_update();
 		    il_get_contents(&cmdln);
 
 		    current_mode = GIT_TERMINAL_MODE;
-		    ks  = tty_get_key(&repeat_count);
-		    key = ks->key_seq[0];
 
+		    while ((ks = tty_get_key(&repeat_count)) == NULL)
+			report_undefined_key(CommandLineModeHelp);
+
+		    key = ks->key_seq[0];
 		    command = (command_t *)ks->aux_data;
+
 		    if (command && command->builtin)
 			key = - 1 - (command->name - builtin[0]) /
 				     MAX_BUILTIN_NAME;
@@ -2402,14 +2470,16 @@ main(argc, argv)
 				tty_put_screen(screen);
 				il_kill_line(IL_DONT_STORE);
 				il_insert_text(output_string);
+				tty_update_title(output_string);
 				start(output_string, 0);
 				tty_get_screen(screen);
 				il_history(IL_RECORD);
 				status(CommandLineModeHelp,
-				       MSG_OK, MSG_CENTERED);
+				       STATUS_OK, STATUS_CENTERED);
 				il_kill_line(IL_DONT_STORE);
 				saved_il = il_save();
 				tty_update_title(panel_get_path(src_panel));
+				tty_update();
 			    }
 
 			    break;
@@ -2419,7 +2489,10 @@ main(argc, argv)
 			    il_free(saved_il);
 
 			    for (i = 0; i < repeat_count; i++)
+			    {
 				il_history(IL_PREVIOUS);
+				tty_update();
+			    }
 
 			    saved_il = il_save();
 			    break;
@@ -2429,7 +2502,10 @@ main(argc, argv)
 			    il_free(saved_il);
 
 			    for (i = 0; i < repeat_count; i++)
+			    {
 				il_history(IL_NEXT);
+				tty_update();
+			    }
 
 			    saved_il = il_save();
 			    break;
@@ -2437,7 +2513,9 @@ main(argc, argv)
 			case BUILTIN_refresh:
 			    refresh(0);
 			    tty_put_screen(screen);
-			    status(CommandLineModeHelp, MSG_OK, MSG_CENTERED);
+			    status(CommandLineModeHelp,
+				   STATUS_OK, STATUS_CENTERED);
+			    tty_update();
 			    break;
 
 			case BUILTIN_exit:
@@ -2448,7 +2526,9 @@ main(argc, argv)
 				goto end_tty_mode;
 			    }
 
-			    status(CommandLineModeHelp, MSG_OK, MSG_CENTERED);
+			    status(CommandLineModeHelp,
+				   STATUS_OK, STATUS_CENTERED);
+			    tty_update();
 			    break;
 
 			default:
@@ -2461,15 +2541,19 @@ main(argc, argv)
 
 				saved_il = il_save();
 			    }
+
 			    break;
 		    }
+
+		    status_update();
 		}
 
 	      end_tty_mode:
 		panel_no_optimizations(src_panel);
 		panel_no_optimizations(dst_panel);
 		tty_touch();
-		status_reset();
+		status_default();
+		tty_update();
 		alarm(60 - get_local_time()->tm_sec);
 
 		current_mode = GIT_SCREEN_MODE;
@@ -2533,25 +2617,48 @@ main(argc, argv)
 		break;
 
 	    case BUILTIN_entry_to_input_line:
-		len = strlen(cmdln);
 		srcptr = panel_get_current_file_name(src_panel);
-		ptr = xmalloc(ptrlen = 1 + 1 + strlen(srcptr) + 1 + 1 + 1);
+		ptr = xmalloc(1 + 1 + strlen(srcptr) + 1 + 1 + 1);
 
 	      copy_to_cmdln:
+		len = strlen(cmdln);
 		il_free(saved_il);
-		sprintf(ptr, " \"%s\" ", srcptr);
-		cmdln = xrealloc(cmdln, len + ptrlen);
+		/* FIXME: I removed the quotes since git is not able to
+		   handle non-printable characters in the command line
+		   anyway.  We just convert them to question marks.  */
+
+		/* If the character before the position where we want
+		   to insert the file name is ' ' or '/', we don't
+		   insert a space before the file name, to avoid
+		   redundant spaces and allow for easy path
+		   concatenation.  */
+		if ((len != 0) &&
+		    ((cmdln[il_point() - 1] == '/') ||
+		     (cmdln[il_point() - 1] == ' ')))
+		{
+		    if (needs_quotes(srcptr, strlen(srcptr)))
+			sprintf(ptr, "\"%s\" ", srcptr);
+		    else
+			sprintf(ptr, "%s ", srcptr);
+		}
+		else
+		{
+		    if (needs_quotes(srcptr, strlen(srcptr)))
+			sprintf(ptr, " \"%s\" ", srcptr);
+		    else
+			sprintf(ptr, " %s ", srcptr);
+		}
+
+		ptrlen = strlen(ptr);
 		toprintable(ptr, strlen(ptr));
-		strcpy(&cmdln[len], ptr);
 		il_insert_text(ptr);
 		xfree(ptr);
 		saved_il = il_save();
 		break;
 
 	    case BUILTIN_other_path_to_input_line:
-		ptrlen = 1 + 1 + dst_panel->pathlen + 1 + 1 + 1;
-		ptr = xmalloc(ptrlen);
 		srcptr = dst_panel->path;
+		ptr = xmalloc(1 + 1 + dst_panel->pathlen + 1 + 1 + 1);
 		goto copy_to_cmdln;
 
 	    case BUILTIN_selected_entries_to_input_line:
@@ -2563,24 +2670,21 @@ main(argc, argv)
 		while ((entry = panel_get_next(src_panel)) != -1)
 		{
 		    srcptr = src_panel->dir_entry[entry].name;
-		    ptr = xmalloc(ptrlen = 1 + 1 + strlen(srcptr) + 1 + 1 + 1);
-		    sprintf(ptr, " \"%s\" ", srcptr);
-		    cmdln = xrealloc(cmdln, len + ptrlen);
+		    ptr = xmalloc(1 + 1 + strlen(srcptr) + 1 + 1 + 1);
 
-		    for (i = 0; ptr[i]; i++)
-			if (!isprint(ptr[i]))
-			{
-			    tty_beep();
-			    ptr[i] = 0;
-			    saved_il = il_save();
-			    break;
-			}
+		    if (needs_quotes(srcptr, strlen(srcptr)))
+			sprintf(ptr, " \"%s\"", srcptr);
+		    else
+			sprintf(ptr, " %s", srcptr);
 
-		    strcpy(&cmdln[len], ptr);
+		    ptrlen = strlen(ptr);
+		    len += ptrlen;
+		    toprintable(ptr, ptrlen);
 		    il_insert_text(ptr);
 		    xfree(ptr);
 		}
 
+		il_insert_text(" ");
 		saved_il = il_save();
 		break;
 
@@ -2588,7 +2692,10 @@ main(argc, argv)
 		il_free(saved_il);
 
 		for (i = 0; i < repeat_count; i++)
+		{
 		    il_history(IL_PREVIOUS);
+		    tty_update();
+		}
 
 		saved_il = il_save();
 		break;
@@ -2597,7 +2704,10 @@ main(argc, argv)
 		il_free(saved_il);
 
 		for (i = 0; i < repeat_count; i++)
+		{
 		    il_history(IL_NEXT);
+		    tty_update();
+		}
 
 		saved_il = il_save();
 		break;
@@ -2711,6 +2821,7 @@ main(argc, argv)
 
 	    case BUILTIN_isearch_backward:
 		previous_isearch_failed = 0;
+		resuming_previous_isearch = 0;
 		il_isearch("I-search backward: ", (char **)NULL,
 			   IL_ISEARCH_BEGIN, (int *)NULL);
 		panel_action(src_panel, act_ISEARCH_BEGIN,
@@ -2724,11 +2835,24 @@ main(argc, argv)
 				   &iai.action) == NULL)
 			break;
 
+		    /* If the strlen(input) == 0, we typed ^R twice,
+		       so we must search for the string we searched
+		       previously.  */
+		    if (strlen(input) == 0 &&
+			search_string && strlen(search_string))
+		    {
+			xfree(input);
+			input = strdup(search_string);
+			il_insert_text(input);
+			resuming_previous_isearch = 1;
+			previous_isearch_failed = 0;
+		    }
+
 		    /* Wrap around.  */
 		    if (iai.action == IL_ISEARCH_ACTION_RETRY &&
 			previous_isearch_failed)
 		    {
-			panel_set_wrapped_isearch_flag(src_panel);
+			panel_set_wrapped_isearch_flag(src_panel, 1);
 			previous_isearch_failed = 0;
 		    }
 
@@ -2742,13 +2866,35 @@ main(argc, argv)
 			tty_beep();
 		    }
 		    else
+		    {
+			int update = 0;
+
+			if (resuming_previous_isearch)
+			    if (iai.length < strlen(search_string))
+			    {
+				il_kill_line(0);
+				resuming_previous_isearch = 0;
+				update = 1;
+			    }
+
 			if (iai.length < strlen(input))
 			{
 			    il_backward_delete_char();
-			    il_full_update();
+			    update = 1;
 			}
+
+			if (update)
+			{
+			    il_update();
+			    il_update_point();
+			    tty_update();
+			}
+		    }
 		}
 
+		if (search_string)
+		    xfree(search_string);
+		search_string = strdup(input);
 		panel_action(src_panel, act_ISEARCH_END, dst_panel, NULL, 1);
 		il_isearch((char *)NULL, (char **)NULL,
 			   IL_ISEARCH_END, (int *)NULL);
@@ -2756,6 +2902,7 @@ main(argc, argv)
 
 	    case BUILTIN_isearch_forward:
 		previous_isearch_failed = 0;
+		resuming_previous_isearch = 0;
 		il_isearch("I-search: ", (char **)NULL,
 			   IL_ISEARCH_BEGIN, (int *)NULL);
 		panel_action(src_panel, act_ISEARCH_BEGIN, dst_panel, NULL, 1);
@@ -2768,11 +2915,25 @@ main(argc, argv)
 				   &iai.action) == NULL)
 			break;
 
+		    /* If the strlen(input) == 0, we typed ^S twice,
+		       so we must search for the string we searched
+		       previously.  */
+		    if (strlen(input) == 0 &&
+			search_string && strlen(search_string))
+		    {
+			xfree(input);
+			input = strdup(search_string);
+			il_insert_text(input);
+			resuming_previous_isearch = 1;
+			previous_isearch_failed = 0;
+		    }
+
 		    /* Wrap around.  */
 		    if (iai.action == IL_ISEARCH_ACTION_RETRY &&
 			previous_isearch_failed)
 		    {
-			panel_set_wrapped_isearch_flag(src_panel);
+			tty_beep();
+			panel_set_wrapped_isearch_flag(src_panel, 1);
 			previous_isearch_failed = 0;
 		    }
 
@@ -2786,13 +2947,35 @@ main(argc, argv)
 			tty_beep();
 		    }
 		    else
+		    {
+			int update = 0;
+
+			if (resuming_previous_isearch)
+			    if (iai.length < strlen(search_string))
+			    {
+				il_kill_line(0);
+				resuming_previous_isearch = 0;
+				update = 1;
+			    }
+
 			if (iai.length < strlen(input))
 			{
 			    il_backward_delete_char();
-			    il_full_update();
+			    update = 1;
 			}
+
+			if (update)
+			{
+			    il_update();
+			    il_update_point();
+			    tty_update();
+			}
+		    }
 		}
 
+		if (search_string)
+		    xfree(search_string);
+		search_string = strdup(input);
 		panel_action(src_panel, act_ISEARCH_END, dst_panel, NULL, 1);
 		il_isearch((char *)NULL, (char **)NULL,
 			   IL_ISEARCH_END, (int *)NULL);
@@ -2887,8 +3070,7 @@ main(argc, argv)
 		    il_read_line("Enter password to unlock: ",
 				 &unlock_password, (char *)NULL,
 				 (xstack_t *)NULL);
-		    il_message(lock_think);
-		    sleep(1);
+		    tty_update();
 
 		    if (unlock_password &&
 			strcmp(lock_password, unlock_password) == 0)
@@ -2896,11 +3078,11 @@ main(argc, argv)
 
 		    il_message(lock_bad);
 		    tty_beep();
+		    tty_update();
 		    sleep(2);
 		}
 
-		il_message(lock_ok);
-		sleep(1);
+		tty_update();
 		xfree(lock_password);
 		xfree(unlock_password);
 
@@ -2962,6 +3144,16 @@ main(argc, argv)
 			     dst_panel, input, repeat_count);
 		break;
 
+	    case BUILTIN_select_extension:
+		panel_action(src_panel, act_SELECT_EXTENSION, dst_panel,
+			     NULL, 1);
+		break;
+
+	    case BUILTIN_unselect_extension:
+		panel_action(src_panel, act_UNSELECT_EXTENSION, dst_panel,
+			     NULL, 1);
+		break;
+
 	    default:
 		if (key)
 		{
