@@ -1,6 +1,6 @@
 /* gitps.c -- A process viewer/killer utility.  */
 
-/* Copyright (C) 1993-1998 Free Software Foundation, Inc.
+/* Copyright (C) 1993-1999 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -17,7 +17,7 @@
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
 /* Written by Tudor Hulubei and Andrei Pitis.  */
-/* $Id: gitps.c,v 1.8 1998/03/13 16:04:00 tudor Exp $ */
+/* $Id: gitps.c,v 1.18 1999/05/31 21:21:04 tudor Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include <config.h>
@@ -166,122 +166,93 @@ int signal_type = 0;
 
 static xsignal_t sigdesc[] =
 {
-    { "SIGTERM  ", SIGTERM   },
-    { "SIGABRT  ", SIGABRT   },
     { "SIGALRM  ", SIGALRM   },
-
+    { "SIGABRT  ", SIGABRT   },
 #ifdef SIGBUS
     { "SIGBUS   ", SIGBUS    },
 #endif
-
 #ifdef SIGCHLD
     { "SIGCHLD  ", SIGCHLD   },
 #endif
-
 #ifdef SIGCLD
     { "SIGCLD   ", SIGCLD    },
 #endif
-
 #ifdef SIGCONT
     { "SIGCONT  ", SIGCONT   },
 #endif
-
 #ifdef SIGEMT
     { "SIGEMT   ", SIGEMT    },
 #endif
-
 #ifdef SIGFPE
     { "SIGFPE   ", SIGFPE    },
 #endif
-
     { "SIGHUP   ", SIGHUP    },
-
 #ifdef SIGILL
     { "SIGILL   ", SIGILL    },
 #endif
-
 #ifdef SIGINFO
     { "SIGINFO  ", SIGINFO   },
 #endif
-
     { "SIGINT   ", SIGINT    },
-
 #ifdef SIGIO
     { "SIGIO    ", SIGIO     },
 #endif
-
 #ifdef SIGIOT
     { "SIGIOT   ", SIGIOT    },
 #endif
-
     { "SIGKILL  ", SIGKILL   },
     { "SIGPIPE  ", SIGPIPE   },
-
 #ifdef SIGPOLL
     { "SIGPOLL  ", SIGPOLL   },
 #endif
-
 #ifdef SIGPROF
     { "SIGPROF  ", SIGPROF   },
 #endif
-
 #ifdef SIGPWR
     { "SIGPWR   ", SIGPWR    },
 #endif
-
     { "SIGQUIT  ", SIGQUIT   },
     { "SIGSEGV  ", SIGSEGV   },
-
 #ifdef SIGSTOP
     { "SIGSTOP  ", SIGSTOP   },
 #endif
-
 #ifdef SIGSYS
     { "SIGSYS   ", SIGSYS    },
 #endif
-
+    { "SIGTERM  ", SIGTERM   },
 #ifdef SIGTRAP
     { "SIGTRAP  ", SIGTRAP   },
 #endif
-
 #ifdef SIGTSTP
     { "SIGTSTP  ", SIGTSTP   },
 #endif
-
 #ifdef SIGTTIN
     { "SIGTTIN  ", SIGTTIN   },
 #endif
-
 #ifdef SIGTTOU
     { "SIGTTOU  ", SIGTTOU   },
 #endif
-
 #ifdef SIGURG
     { "SIGURG   ", SIGURG    },
 #endif
-
     { "SIGUSR1  ", SIGUSR1   },
     { "SIGUSR2  ", SIGUSR2   },
-
 #ifdef SIGVTALRM
     { "SIGVTALRM", SIGVTALRM },
 #endif
-
 #ifdef SIGWINCH
     { "SIGWINCH ", SIGWINCH  },
 #endif
-
 #ifdef SIGXCPU
     { "SIGXCPU  ", SIGXCPU   },
 #endif
-
 #ifdef SIGXFSZ
     { "SIGXFSZ  ", SIGXFSZ   },
 #endif
 };
 
 
-#define BUILTIN_OPERATIONS              26
+#define BUILTIN_OPERATIONS              46
 
 
 #define BUILTIN_previous_line            0
@@ -291,24 +262,45 @@ static xsignal_t sigdesc[] =
 #define BUILTIN_beginning_of_list        4
 #define BUILTIN_end_of_list              5
 #define BUILTIN_next_signal              6
-#define BUILTIN_SIGHUP                   7
-#define BUILTIN_SIGINT                   8
-#define BUILTIN_SIGQUIT                  9
-#define BUILTIN_SIGILL                  10
-#define BUILTIN_SIGFPE                  11
-#define BUILTIN_SIGKILL                 12
-#define BUILTIN_SIGUSR1                 13
-#define BUILTIN_SIGSEGV                 14
-#define BUILTIN_SIGUSR2                 15
-#define BUILTIN_SIGPIPE                 16
-#define BUILTIN_SIGALRM                 17
-#define BUILTIN_SIGTERM                 18
-#define BUILTIN_SIGCHLD                 19
-#define BUILTIN_SIGCONT                 20
-#define BUILTIN_kill_process            21
-#define BUILTIN_refresh                 22
-#define BUILTIN_exit                    23
-#define BUILTIN_hard_refresh            24
+#define BUILTIN_SIGALRM                  7
+#define BUILTIN_SIGABRT                  8
+#define BUILTIN_SIGBUS                   9
+#define BUILTIN_SIGCHLD                 10
+#define BUILTIN_SIGCLD                  11
+#define BUILTIN_SIGCONT                 12
+#define BUILTIN_SIGEMT                  13
+#define BUILTIN_SIGFPE                  14
+#define BUILTIN_SIGHUP                  15
+#define BUILTIN_SIGILL                  16
+#define BUILTIN_SIGINFO                 17
+#define BUILTIN_SIGINT                  18
+#define BUILTIN_SIGIO                   19
+#define BUILTIN_SIGIOT                  20
+#define BUILTIN_SIGKILL                 21
+#define BUILTIN_SIGPIPE                 22
+#define BUILTIN_SIGPOLL                 23
+#define BUILTIN_SIGPROF                 24
+#define BUILTIN_SIGPWR                  25
+#define BUILTIN_SIGQUIT                 26
+#define BUILTIN_SIGSEGV                 27
+#define BUILTIN_SIGSTOP                 28
+#define BUILTIN_SIGSYS                  29
+#define BUILTIN_SIGTERM                 30
+#define BUILTIN_SIGTRAP                 31
+#define BUILTIN_SIGTSTP                 32
+#define BUILTIN_SIGTTIN                 33
+#define BUILTIN_SIGTTOU                 34
+#define BUILTIN_SIGURG                  35
+#define BUILTIN_SIGUSR1                 36
+#define BUILTIN_SIGUSR2                 37
+#define BUILTIN_SIGVTALRM               38
+#define BUILTIN_SIGWINCH                39
+#define BUILTIN_SIGXCPU                 40
+#define BUILTIN_SIGXFSZ                 41
+#define BUILTIN_kill_process            42
+#define BUILTIN_refresh                 43
+#define BUILTIN_exit                    44
+#define BUILTIN_hard_refresh            45
 
 
 #define MAX_BUILTIN_NAME                20
@@ -322,20 +314,41 @@ char built_in[BUILTIN_OPERATIONS][MAX_BUILTIN_NAME] =
     "beginning-of-list",
     "end-of-list",
     "next-signal",
-    "SIGHUP",
-    "SIGINT",
-    "SIGQUIT",
-    "SIGILL",
-    "SIGFPE",
-    "SIGKILL",
-    "SIGUSR1",
-    "SIGSEGV",
-    "SIGUSR2",
-    "SIGPIPE",
     "SIGALRM",
-    "SIGTERM",
+    "SIGABRT",
+    "SIGBUS",
     "SIGCHLD",
+    "SIGCLD",
     "SIGCONT",
+    "SIGEMT",
+    "SIGFPE",
+    "SIGHUP",
+    "SIGILL",
+    "SIGINFO",
+    "SIGINT",
+    "SIGIO",
+    "SIGIOT",
+    "SIGKILL",
+    "SIGPIPE",
+    "SIGPOLL",
+    "SIGPROF",
+    "SIGPWR",
+    "SIGQUIT",
+    "SIGSEGV",
+    "SIGSTOP",
+    "SIGSYS",
+    "SIGTERM",
+    "SIGTRAP",
+    "SIGTSTP",
+    "SIGTTIN",
+    "SIGTTOU",
+    "SIGURG",
+    "SIGUSR1",
+    "SIGUSR2",
+    "SIGVTALRM",
+    "SIGWINCH",
+    "SIGXCPU",
+    "SIGXFSZ",
     "kill-process",
     "refresh",
     "exit",
@@ -409,24 +422,63 @@ set_status(what)
 
 
 void
-set_signal()
+set_signal(int index)
 {
-    int len = sizeof(sigdesc[0].signame) - 1;
+    int i, len = sizeof(sigdesc[0].signame) - 1;
+
+    if (index >= 0)
+	for (i = 0; i < (int)(sizeof(sigdesc) / sizeof(xsignal_t)); i++)
+	    if (sigdesc[i].signal == index)
+	    {
+		signal_type = i;
+		break;
+	    }
 
     if (tty_columns > len)
     {
 	tty_colors(StatusBrightness, WHITE, StatusBackground);
 	window_goto(status_window, 0, tty_columns - len - 1);
 	window_puts(status_window, sigdesc[signal_type].signame, len);
-
-	if (UseLastScreenChar == ON)
-	    window_putc(status_window, ' ');
+	window_putc(status_window, ' ');
     }
 
     window_goto(status_window, 0, tty_columns - 1);
 }
 
 
+void
+report_undefined_key()
+{
+    char *prev = tty_get_previous_key_seq();
+    size_t length = strlen(prev);
+
+    if (length && (prev[length - 1] != key_INTERRUPT))
+    {
+	char *str = (char *)tty_key_machine2human(tty_get_previous_key_seq());
+	char *buf = xmalloc(128 + strlen(str));
+
+	sprintf(buf, "%s: not defined.", str);
+	memset(global_buf, ' ', tty_columns);
+	memcpy(global_buf, buf, min(tty_columns, (int)strlen(buf)));
+	xfree(buf);
+
+	tty_colors(ON, WHITE, RED);
+	window_goto(status_window, 0, 0);
+	window_puts(status_window, global_buf, tty_columns);
+
+	tty_beep();
+	tty_update();
+	sleep(1);
+    }
+    else
+	tty_beep();
+
+    set_status((char *)NULL);
+    set_signal(-1);
+    tty_update();
+}
+
+
 void
 free_ps_list()
 {
@@ -479,13 +531,13 @@ get_PID_index(ps_output)
 
     for (i = 0; ; i++)
     {
-	while (isspace(*h))
+	while (isspace((int)*h))
 	    h++;
 
 	if (memcmp(h, "PID", 3) == 0)
 	    return i;
 
-	while (!isspace(*h))
+	while (!isspace((int)*h))
 	    h++;
     }
 }
@@ -505,22 +557,22 @@ kill_process(process_index)
 
     for (i = 0; i < PID_index; i++)
     {
-	while (isspace(*p))
+	while (isspace((int)*p))
 	    p++;
 
 	if (memcmp(p, "PID", 3) == 0)
 	    return i;
 
-	while (!isspace(*p))
+	while (!isspace((int)*p))
 	    p++;
     }
 
     i = 0;
 
-    while (isspace(*p))
+    while (isspace((int)*p))
 	p++;
 
-    while (!isspace(*p))
+    while (!isspace((int)*p))
 	pidstr[i++] = *p++;
 
     pidstr[i] = 0;
@@ -554,7 +606,7 @@ update_process(process, update_color)
 
     memset(global_buf, ' ', tty_columns);
     memcpy(global_buf, ps_vect[process],
-	   min(tty_columns, (int)strlen(ps_vect[process]) - 1));
+	   min(tty_columns, (int)strlen(ps_vect[process])));
 
     if (update_color)
     {
@@ -737,8 +789,8 @@ read_keys(keys)
 	    {
 		/* This is not a TERMCAP symbol, it is a key sequence
 		   that we will have to convert it with
-		   tty_key_convert() into a machine usable form before
-		   using it.  */
+		   tty_key_human2machine() into a machine usable form
+		   before using it.  */
 		need_conversion = 1;
 	    }
 	}
@@ -754,7 +806,8 @@ read_keys(keys)
 
 	if (j < BUILTIN_OPERATIONS)
 	{
-	    if (!need_conversion || tty_key_convert((unsigned char *)key_seq))
+	    if (!need_conversion ||
+		tty_key_human2machine((unsigned char *)key_seq))
 		tty_key_list_insert((unsigned char *)key_seq, built_in[j]);
 	}
 	else
@@ -861,7 +914,7 @@ refresh(signum)
     set_title();
     set_header();
     set_status((char *)NULL);
-    set_signal();
+    set_signal(SIGTERM);
     update_all();
     tty_update();
 
@@ -891,7 +944,6 @@ usage()
     printf("usage: %s [-hvilcbp]\n", program);
     printf(" -h         print this help message\n");
     printf(" -v         print the version number\n");
-    printf(" -i         print the installation directory\n");
     printf(" -c         use ANSI colors\n");
     printf(" -b         don't use ANSI colors\n");
     printf(" -l         don't use the last screen character\n");
@@ -932,7 +984,7 @@ main(argc, argv)
 	ansi_colors = ON;
 
     /* Parse the command line.  */
-    while ((c = getopt(argc, argv, "hvicblp")) != -1)
+    while ((c = getopt(argc, argv, "hvcblp")) != -1)
 	switch (c)
 	{
 	    case 'h':
@@ -945,11 +997,6 @@ main(argc, argv)
 		printf("%s %s\n", PRODUCT, VERSION);
 		return 0;
 
-	    case 'i':
-		/* Installation directory request.  */
-		printf("%s is installed in %s.\n", PRODUCT, PREFIX);
-		return 0;
-
 	    case 'c':
 		/* Force git to use ANSI color sequences.  */
 		ansi_colors = ON;
@@ -1023,22 +1070,15 @@ main(argc, argv)
 	UseLastScreenChar = OFF;
 
     tty_set_last_char_flag(UseLastScreenChar);
-
     StartupScrollStep  = get_int_var("StartupScrollStep", (tty_lines - 3) / 2);
 
-
     use_section("[GITPS-Setup]");
-
     help = get_string_var("Help", "");
 
-
     use_section(AnsiColors ? color_section : monochrome_section);
-
     get_colorset_var(PSColors, PSFields, PS_FIELDS);
 
-
     use_section("[GITPS-Keys]");
-
     keys = read_keys(keys);
 
     if (keys == MAX_KEYS)
@@ -1103,7 +1143,12 @@ main(argc, argv)
 
     while (1)
     {
-	ks  = tty_get_key(&repeat_count);
+	while ((ks = tty_get_key(&repeat_count)) == NULL)
+	    report_undefined_key();
+
+	set_status((char *)NULL);
+	set_signal(-1);
+
 	key = ((char *)ks->aux_data - (char *)built_in) / MAX_BUILTIN_NAME;
 
 	switch (key)
@@ -1249,23 +1294,92 @@ main(argc, argv)
 	    case BUILTIN_next_signal:
 		signal_type++;
 		signal_type %= sizeof(sigdesc) / sizeof(xsignal_t);
-		set_signal();
+		set_signal(-1);
 		break;
 
-	    case BUILTIN_SIGHUP:  signal_type =  0; set_signal(); break;
-	    case BUILTIN_SIGINT:  signal_type =  1; set_signal(); break;
-	    case BUILTIN_SIGQUIT: signal_type =  2; set_signal(); break;
-	    case BUILTIN_SIGILL:  signal_type =  3; set_signal(); break;
-	    case BUILTIN_SIGFPE:  signal_type =  4; set_signal(); break;
-	    case BUILTIN_SIGKILL: signal_type =  5; set_signal(); break;
-	    case BUILTIN_SIGUSR1: signal_type =  6; set_signal(); break;
-	    case BUILTIN_SIGSEGV: signal_type =  7; set_signal(); break;
-	    case BUILTIN_SIGUSR2: signal_type =  8; set_signal(); break;
-	    case BUILTIN_SIGPIPE: signal_type =  9; set_signal(); break;
-	    case BUILTIN_SIGALRM: signal_type = 10; set_signal(); break;
-	    case BUILTIN_SIGTERM: signal_type = 11; set_signal(); break;
-	    case BUILTIN_SIGCHLD: signal_type = 12; set_signal(); break;
-	    case BUILTIN_SIGCONT: signal_type = 13; set_signal(); break;
+	    case BUILTIN_SIGALRM:   set_signal(SIGALRM);   break;
+	    case BUILTIN_SIGABRT:   set_signal(SIGABRT);   break;
+#ifdef SIGBUS
+	    case BUILTIN_SIGBUS:    set_signal(SIGBUS);    break;
+#endif
+#ifdef SIGCHLD
+	    case BUILTIN_SIGCHLD:   set_signal(SIGCHLD);   break;
+#endif
+#ifdef SIGCLD
+	    case BUILTIN_SIGCLD:    set_signal(SIGCLD);    break;
+#endif
+#ifdef SIGCONT
+	    case BUILTIN_SIGCONT:   set_signal(SIGCONT);   break;
+#endif
+#ifdef SIGEMT
+	    case BUILTIN_SIGEMT:    set_signal(SIGEMT);    break;
+#endif
+#ifdef SIGFPE
+	    case BUILTIN_SIGFPE:    set_signal(SIGFPE);    break;
+#endif
+	    case BUILTIN_SIGHUP:    set_signal(SIGHUP);    break;
+#ifdef SIGILL
+	    case BUILTIN_SIGILL:    set_signal(SIGILL);    break;
+#endif
+#ifdef SIGINFO
+	    case BUILTIN_SIGINFO:   set_signal(SIGINFO);   break;
+#endif
+	    case BUILTIN_SIGINT:    set_signal(SIGINT);    break;
+#ifdef SIGIO
+	    case BUILTIN_SIGIO:     set_signal(SIGIO);     break;
+#endif
+#ifdef SIGIOT
+	    case BUILTIN_SIGIOT:    set_signal(SIGIOT);    break;
+#endif
+	    case BUILTIN_SIGKILL:   set_signal(SIGKILL);   break;
+	    case BUILTIN_SIGPIPE:   set_signal(SIGPIPE);   break;
+#ifdef SIGPOLL
+	    case BUILTIN_SIGPOLL:   set_signal(SIGPOLL);   break;
+#endif
+#ifdef SIGPROF
+	    case BUILTIN_SIGPROF:   set_signal(SIGPROF);   break;
+#endif
+#ifdef SIGPWR
+	    case BUILTIN_SIGPWR:    set_signal(SIGPWR);    break;
+#endif
+	    case BUILTIN_SIGQUIT:   set_signal(SIGQUIT);   break;
+	    case BUILTIN_SIGSEGV:   set_signal(SIGSEGV);   break;
+#ifdef SIGSTOP
+	    case BUILTIN_SIGSTOP:   set_signal(SIGSTOP);   break;
+#endif
+#ifdef SIGSYS
+	    case BUILTIN_SIGSYS:    set_signal(SIGSYS);    break;
+#endif
+	    case BUILTIN_SIGTERM:   set_signal(SIGTERM);   break;
+#ifdef SIGTRAP
+	    case BUILTIN_SIGTRAP:   set_signal(SIGTRAP);   break;
+#endif
+#ifdef SIGTSTP
+	    case BUILTIN_SIGTSTP:   set_signal(SIGTSTP);   break;
+#endif
+#ifdef SIGTTIN
+	    case BUILTIN_SIGTTIN:   set_signal(SIGTTIN);   break;
+#endif
+#ifdef SIGTTOU
+	    case BUILTIN_SIGTTOU:   set_signal(SIGTTOU);   break;
+#endif
+#ifdef SIGURG
+	    case BUILTIN_SIGURG:    set_signal(SIGURG);    break;
+#endif
+	    case BUILTIN_SIGUSR1:   set_signal(SIGUSR1);   break;
+	    case BUILTIN_SIGUSR2:   set_signal(SIGUSR2);   break;
+#ifdef SIGVTALRM
+	    case BUILTIN_SIGVTALRM: set_signal(SIGVTALRM); break;
+#endif
+#ifdef SIGWINCH
+	    case BUILTIN_SIGWINCH:  set_signal(SIGWINCH);  break;
+#endif
+#ifdef SIGXCPU
+	    case BUILTIN_SIGXCPU:   set_signal(SIGXCPU);   break;
+#endif
+#ifdef SIGXFSZ
+	    case BUILTIN_SIGXFSZ:   set_signal(SIGXFSZ);   break;
+#endif
 
 	    case BUILTIN_hard_refresh:
 		tty_touch();
@@ -1287,13 +1401,18 @@ main(argc, argv)
 		    tmp = xmalloc(16 + strlen((e == EPERM) ? no_perm:no_proc));
 		    sprintf(tmp, "Error: %s", (e == EPERM) ? no_perm:no_proc);
 		    set_status(tmp);
+		    tty_update();
 		    xfree(tmp);
 		    errno = 0;
 		    tty_get_key(NULL);
 		    set_status((char *)NULL);
-		    set_signal();
+		    set_signal(-1);
 		}
 		break;
+
+	    default:
+		report_undefined_key();
+		break;
 	}
     }
 
