@@ -1,6 +1,6 @@
 /* gitview.c -- A hex/ascii file viewer.  */
 
-/* Copyright (C) 1993-1998 Free Software Foundation, Inc.
+/* Copyright (C) 1993-1999 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -17,7 +17,9 @@
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
 /* Written by Tudor Hulubei and Andrei Pitis.  */
-/* $Id: gitview.c,v 1.5 1998/03/10 05:12:12 tudor Exp $ */
+/* Patched by Gregory Gerard <ggerard@ssl.sel.sony.com> to display on
+   the entire screen.  */
+/* $Id: gitview.c,v 1.14 1999/05/31 21:21:04 tudor Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include <config.h>
@@ -46,6 +48,8 @@
 #include <unistd.h>
 #endif /* HAVE_UNISTD_H */
 
+#include <assert.h>
+
 #include "stdc.h"
 #include "xstring.h"
 #include "xmalloc.h"
@@ -150,6 +154,10 @@ int AnsiColors = OFF;
 #endif  /* !HAVE_LINUX */
 
 
+#define VIEW_LINES (tty_lines - 9)
+#define SEEK_LINE (tty_lines - 5)
+
+
 #ifdef STAT_MACROS_BROKEN
 #ifdef S_IFREG
 #undef S_IFREG
@@ -183,7 +191,7 @@ char *home;
 char *program;
 char *screen;
 char *filename;
-int cnt, size;
+int count, size;
 char offset[16];
 char *header_text;
 int  UseLastScreenChar;
@@ -219,6 +227,16 @@ file_length()
 }
 
 
+void
+cursor_update()
+{
+    if (tty_lines >= 9)
+	window_goto(file_window, SEEK_LINE, strlen(seek_txt) + count);
+    else
+	window_goto(file_window, tty_lines - 1, tty_columns - 1);
+}
+
+
 void
 set_title()
 {
@@ -256,11 +274,39 @@ set_status()
     tty_colors(StatusBrightness, StatusForeground, StatusBackground);
 
     window_goto(status_window, 0, 0);
+    window_puts(status_window, global_buf, tty_columns);
+}
 
-    if (UseLastScreenChar)
+
+void
+report_undefined_key()
+{
+    char *prev = tty_get_previous_key_seq();
+    size_t length = strlen(prev);
+
+    if (length && (prev[length - 1] != key_INTERRUPT))
+    {
+	char *str = (char *)tty_key_machine2human(tty_get_previous_key_seq());
+	char *buf = xmalloc(128 + strlen(str));
+	sprintf(buf, "%s: not defined.", str);
+	memset(global_buf, ' ', tty_columns);
+	memcpy(global_buf, buf, min(tty_columns, (int)strlen(buf)));
+	xfree(buf);
+
+	tty_colors(ON, WHITE, RED);
+	window_goto(status_window, 0, 0);
 	window_puts(status_window, global_buf, tty_columns);
+
+	tty_beep();
+	tty_update();
+	sleep(1);
+    }
     else
-	window_puts(status_window, global_buf, tty_columns - 1);
+	tty_beep();
+
+    set_status();
+    cursor_update();
+    tty_update();
 }
 
 
@@ -268,7 +314,7 @@ char
 char_to_print(c)
     char c;
 {
-    return isprint(c) ? c : '.';
+    return isprint((int)c) ? c : '.';
 }
 
 
@@ -328,7 +374,7 @@ update_all()
 
     tty_colors(ScreenBrightness, ScreenForeground, ScreenBackground);
 
-    for (i = current_line; i < current_line + 16; i++)
+    for (i = current_line; i < current_line + VIEW_LINES; i++)
     {
 	window_goto(file_window, 3 + i - current_line, 0);
 	update_line(i);
@@ -389,8 +435,8 @@ read_keys(keys)
 	    {
 		/* This is not a TERMCAP symbol, it is a key sequence
 		   that we will have to convert it with
-		   tty_key_convert() into a machine usable form before
-		   using it.  */
+		   tty_key_human2machine() into a machine usable form
+		   before using it.  */
 		need_conversion = 1;
 	    }
 	}
@@ -406,7 +452,8 @@ read_keys(keys)
 
 	if (j < BUILTIN_OPERATIONS)
 	{
-	    if (!need_conversion || tty_key_convert((unsigned char *)key_seq))
+	    if (!need_conversion ||
+		tty_key_human2machine((unsigned char *)key_seq))
 		tty_key_list_insert((unsigned char *)key_seq,  built_in[-j-1]);
 	}
 	else
@@ -422,7 +469,7 @@ void
 resize(resize_required)
     int resize_required;
 {
-    int display_title = OFF;
+    int display_status = OFF;
     int display_header = OFF;
     int display_file = OFF;
     int old_tty_lines = tty_lines;
@@ -447,7 +494,7 @@ resize(resize_required)
      * more line for the input line and one for the status.
      *
      * Cases (lines available):
-     *      1 line:	the status
+     *      1 line:	title
      *      2 lines:	title + status
      *      3 lines:	title + header + status
      *	 >= 4 lines:	everything
@@ -458,10 +505,10 @@ resize(resize_required)
     global_buf  = xrealloc(global_buf, tty_columns + 1);
     header_text = xrealloc(header_text, strlen(filename) + 10);
 
-    window_resize(status_window, 0, tty_lines - 1, 1, tty_columns);
+    window_resize(title_window, 0, 0, 1, tty_columns);
 
     if (tty_lines >= 2)
-	display_title = ON;
+	display_status = ON;
 
     if (tty_lines >= 3)
 	display_header = ON;
@@ -469,8 +516,10 @@ resize(resize_required)
     if (tty_lines >= 4)
 	display_file = ON;
 
-    window_resize(title_window, 0, 0, display_title ? 1 : 0, tty_columns);
-    window_resize(header_window, 0, 1, display_header ? 1 : 0, tty_columns);
+    window_resize(status_window, 0, tty_lines - 1,
+		  display_status ? 1 : 0, tty_columns);
+    window_resize(header_window, 0, 1,
+		  display_header ? 1 : 0, tty_columns);
     window_resize(file_window, 0, 2,
 		  display_file ? (tty_lines - 2) : 0, tty_columns);
 }
@@ -496,26 +545,46 @@ refresh(signum)
     tty_colors(ScreenBrightness, ScreenForeground, ScreenBackground);
     tty_fill();
 
-    window_goto(file_window, 1, 0);
-    window_puts(file_window, info_txt, sizeof(info_txt) - 1);
-    window_goto(file_window, 2, 0);
-    window_puts(file_window, line_txt, sizeof(line_txt) - 1);
+    if (tty_lines >= 5)
+    {
+	window_goto(file_window, 1, 0);
+	window_puts(file_window, info_txt, sizeof(info_txt) - 1);
+    }
+
+    if (tty_lines >= 6)
+    {
+	window_goto(file_window, 2, 0);
+	window_puts(file_window, line_txt, sizeof(line_txt) - 1);
+    }
 
     size = file_length();
     lines = size / 16 + (size % 16 ? 1 : 0);
-    current_line = min(current_line, (lines / 16) * 16);
 
-    window_goto(file_window, 20, 0);
-    window_puts(file_window, seek_txt, sizeof(seek_txt) - 1);
-    window_goto(file_window, 20, sizeof(seek_txt) - 1);
-    window_puts(file_window, offset, cnt);
+    if (tty_lines >= 9)
+    {
+	if (VIEW_LINES == 0)
+	    current_line = 0;
+	else
+	    current_line = min(current_line, (lines/VIEW_LINES)*VIEW_LINES);
+
+	window_goto(file_window, SEEK_LINE, 0);
+	window_puts(file_window, seek_txt, sizeof(seek_txt) - 1);
+	window_goto(file_window, SEEK_LINE, sizeof(seek_txt) - 1);
+	window_puts(file_window, offset, count);
+    }
+    else
+	current_line = 0;
 
     set_title();
-    set_header();
     set_status();
+    set_header();
+
     update_all();
 
-    window_goto(file_window, 20, sizeof(seek_txt) - 1 + cnt);
+    if (tty_lines >= 9)
+	window_goto(file_window, SEEK_LINE, sizeof(seek_txt) - 1 + count);
+    else
+	window_goto(file_window, tty_lines - 1, tty_columns - 1);
 
     tty_update();
 
@@ -545,7 +614,6 @@ usage()
     printf("usage: %s [-hvicbl] file\n", program);
     printf(" -h         print this help message\n");
     printf(" -v         print the version number\n");
-    printf(" -i         print the installation directory\n");
     printf(" -c         use ANSI colors\n");
     printf(" -b         don't use ANSI colors\n");
     printf(" -l         don't use the last screen character\n");
@@ -557,7 +625,7 @@ main(argc, argv)
     int argc;
     char *argv[];
 {
-    int key;
+    int key = 0;
     struct stat s;
     tty_key_t *ks;
     int keys, repeat_count, need_update;
@@ -581,7 +649,7 @@ main(argc, argv)
 	ansi_colors = ON;
 
     /* Parse the command line.  */
-    while ((c = getopt(argc, argv, "hvicblp")) != -1)
+    while ((c = getopt(argc, argv, "hvcblp")) != -1)
 	switch (c)
 	{
 	    case 'h':
@@ -594,11 +662,6 @@ main(argc, argv)
 		printf("%s %s\n", PRODUCT, VERSION);
 		return 0;
 
-	    case 'i':
-		/* Installation directory request.  */
-		printf("%s is installed in %s.\n", PRODUCT, PREFIX);
-		return 0;
-
 	    case 'c':
 		/* Force git to use ANSI color sequences.  */
 		ansi_colors = ON;
@@ -666,7 +729,6 @@ main(argc, argv)
     configuration_end();
 
     specific_configuration_init();
-
     use_section("[Setup]");
 
     if (ansi_colors == -1)
@@ -681,18 +743,13 @@ main(argc, argv)
 
     tty_set_last_char_flag(UseLastScreenChar);
 
-
     use_section("[GITVIEW-Setup]");
-
     help = get_string_var("Help", "");
 
     use_section(AnsiColors ? color_section : monochrome_section);
-
     get_colorset_var(ViewerColors, ViewerFields, VIEWER_FIELDS);
 
-
     use_section("[GITVIEW-Keys]");
-
     keys = read_keys(keys);
 
     if (keys == MAX_KEYS)
@@ -714,11 +771,10 @@ main(argc, argv)
 
     signal_handlers(ON);
 
-    offset[cnt]  = 0;
+    offset[count]  = 0;
     current_line = 0;
 
     sprintf(header_text, " File: %s", filename);
-
     tty_update_title(header_text);
 
   restart:
@@ -726,7 +782,9 @@ main(argc, argv)
 
     while (1)
     {
-	ks  = tty_get_key(&repeat_count);
+	while ((ks = tty_get_key(&repeat_count)) == NULL)
+	    report_undefined_key();
+	set_status();
 
 	if (ks->aux_data == NULL)
 	    key = ks->key_seq[0];
@@ -738,9 +796,6 @@ main(argc, argv)
 
 	switch (key)
 	{
-	    case key_INTERRUPT:
-		goto end;
-
 	    case BUILTIN_previous_line:
 		need_update = 0;
 
@@ -762,7 +817,7 @@ main(argc, argv)
 
 		while (repeat_count--)
 		{
-		    if (current_line >= lines - 16)
+		    if (current_line >= lines - VIEW_LINES)
 			break;
 
 		    current_line++, need_update = 1;
@@ -773,19 +828,40 @@ main(argc, argv)
 
 		break;
 
+	    case BUILTIN_backspace:
+		if (count)
+		{
+		    count--;
+
+		    if (tty_lines >= 9)
+		    {
+			tty_colors(ScreenBrightness,
+				   ScreenForeground,
+				   ScreenBackground);
+			window_goto(file_window, SEEK_LINE,
+				    strlen(seek_txt) + count);
+			window_putc(file_window, ' ');
+		    }
+		    else
+			window_goto(file_window,
+				    tty_lines - 1, tty_columns - 1);
+		    break;
+		}
+
 	    case BUILTIN_scroll_down:
 		if (current_line == 0)
 		    break;
 
-		current_line = max(0, current_line - 16);
+		current_line = max(0, current_line - VIEW_LINES);
 		update_all();
 		break;
 
+	    case ' ':
 	    case BUILTIN_scroll_up:
-		if (current_line >= lines - 16)
+		if (current_line >= lines - VIEW_LINES)
 		    break;
 
-		current_line += 16;
+		current_line = min(lines - 1, current_line + 16);
 		update_all();
 		break;
 
@@ -799,9 +875,9 @@ main(argc, argv)
 		break;
 
 	    case BUILTIN_end_of_file:
-		if (regular_file && current_line < lines - 16)
+		if (regular_file && current_line < lines - VIEW_LINES)
 		{
-		    current_line = lines - 16;
+		    current_line = lines - VIEW_LINES;
 		    update_all();
 		}
 
@@ -815,57 +891,65 @@ main(argc, argv)
 	    case '6': case '7': case '8': case '9': case 'A': case 'B':
 	    case 'C': case 'D': case 'E': case 'F': case 'a': case 'b':
 	    case 'c': case 'd': case 'e': case 'f':
-		if (cnt < 8)
+		if (count < 8)
 		{
-		    char tmp;
-		    window_goto(file_window, 20, strlen(seek_txt) + cnt);
-		    tmp = (char)key;
-		    window_putc(file_window, tmp);
-		    offset[cnt++] = tmp;
+		    if (tty_lines >= 9)
+		    {
+			char tmp;
+			tty_colors(ScreenBrightness,
+				   ScreenForeground,
+				   ScreenBackground);
+			window_goto(file_window, SEEK_LINE,
+				    strlen(seek_txt) + count);
+			tmp = (char)key;
+			window_putc(file_window, tmp);
+			offset[count++] = tmp;
+		    }
 		}
 		else
 		    tty_beep();
 		break;
 
-	    case BUILTIN_backspace:
-		if (cnt)
-		    cnt--;
-
-		window_goto(file_window, 20, strlen(seek_txt) + cnt);
-		window_putc(file_window, ' ');
-		break;
-
 	    case BUILTIN_action:
-		if (cnt == 0)
+		if (count == 0)
 		    tty_beep();
 		else
 		{
-		    offset[cnt] = 0;
-		    sscanf(offset, "%x", &cnt);
-		    window_goto(file_window, 20, strlen(seek_txt));
-		    window_puts(file_window, "        ", 8);
+		    if (tty_lines >= 9)
+		    {
+			offset[count] = 0;
 
-		    if (cnt < 0)
-			cnt = 0;
+			sscanf(offset, "%x", &count);
+			tty_colors(ScreenBrightness,
+				   ScreenForeground,
+				   ScreenBackground);
+			window_goto(file_window, SEEK_LINE, strlen(seek_txt));
+			window_puts(file_window, "        ", 8);
 
-		    if (cnt > size)
-			cnt = size;
+			if (count < 0)
+			    count = 0;
 
-		    current_line = cnt >> 4;
-		    update_all();
-		    cnt = 0;
+			if (count > size)
+			    count = size;
+
+			current_line = count >> 4;
+			update_all();
+			count = 0;
+		    }
 		}
 
 		break;
 
+	    case 'q':
 	    case BUILTIN_exit:
 		goto end;
 
 	    default:
+		report_undefined_key();
 		break;
 	}
 
-	window_goto(file_window, 20, strlen(seek_txt) + cnt);
+	cursor_update();
     }
 
   end:
