@@ -1,6 +1,6 @@
 /* inputline.h -- Prototypes for the functions in inputline.c.  */
 
-/* Copyright (C) 1993-1998 Free Software Foundation, Inc.
+/* Copyright (C) 1993-1999 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -17,7 +17,7 @@
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
 /* Written by Tudor Hulubei and Andrei Pitis.  */
-/* $Id: inputline.h,v 1.3 1998/03/10 05:12:12 tudor Exp $ */
+/* $Id: inputline.h,v 1.5 1999/04/22 02:30:56 tudor Exp $ */
 
 #ifndef _GIT_INPUTLINE_H
 #define _GIT_INPUTLINE_H
@@ -90,6 +90,7 @@ extern void il_free PROTO ((input_line_t *));
 extern void il_resize PROTO ((int, int));
 extern input_line_t *il_save PROTO (());
 extern void il_restore PROTO ((input_line_t *));
+extern size_t il_point PROTO (());
 extern int il_echo PROTO ((int));
 extern int il_is_empty PROTO (());
 extern void il_set_mark PROTO (());
@@ -121,7 +122,6 @@ extern void il_set_static_text PROTO ((char *));
 extern void il_insert_text PROTO ((char *));
 extern void il_update_point PROTO (());
 extern void il_update PROTO (());
-extern void il_full_update PROTO (());
 extern int il_get_contents PROTO ((char **));
 extern void il_message PROTO ((char *));
 extern void il_set_error_flag PROTO ((int));
