@@ -1,6 +1,6 @@
 /* misc.h -- Prototypes for the functions in misc.c.  */
 
-/* Copyright (C) 1993-1998 Free Software Foundation, Inc.
+/* Copyright (C) 1993-1999 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -17,7 +17,7 @@
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
 /* Written by Tudor Hulubei and Andrei Pitis.  */
-/* $Id: misc.h,v 1.3 1998/03/10 05:12:12 tudor Exp $ */
+/* $Id: misc.h,v 1.7 1999/04/20 16:38:51 tudor Exp $ */
 
 #ifndef _GIT_MISC_H
 #define _GIT_MISC_H
@@ -55,6 +55,12 @@ extern char monochrome_section[];
 #define min(a, b) (((a) <= (b)) ? (a) : (b))
 
 
+/* It looks like we need this for SunOS.  */
+#ifndef RAND_MAX
+#define RAND_MAX 2147483647
+#endif
+
+
 #define close2(fd1, fd2)        \
 {                               \
      close(fd1);                \
@@ -104,6 +110,7 @@ extern int is_a_bg_command PROTO ((char *));
 extern int is_an_empty_command PROTO ((char *));
 extern void get_file_type_info PROTO (());
 extern void toprintable PROTO ((char *, size_t));
+extern int needs_quotes PROTO ((char *, size_t));
 extern int xsetenv PROTO ((char *, char *));
 
 
