@@ -1,6 +1,6 @@
 /* tty.c -- The tty management file.  */
 
-/* Copyright (C) 1993-1998 Free Software Foundation, Inc.
+/* Copyright (C) 1993-1999 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -17,7 +17,7 @@
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
 /* Written by Tudor Hulubei and Andrei Pitis.  */
-/* $Id: tty.c,v 1.8 1998/03/14 16:05:16 tudor Exp $ */
+/* $Id: tty.c,v 1.25 1999/06/01 00:04:58 tudor Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include <config.h>
@@ -97,9 +97,9 @@ static int vcs_read_ok;
 static int vcs_is_monochrome;
 #endif
 
-/* If tty_kbdmode == 1, single character key sequences are inserted
-   into the linked list.  This feature is used by gitps which has no
-   command line.  */
+/* If tty_kbdmode == TTY_FULL_INPUT, single character key sequences
+   are inserted into the linked list.  This feature is used by gitps
+   which has no command line.  */
 static int tty_kbdmode;
 
 #ifdef HAVE_POSIX_TTY
@@ -129,16 +129,19 @@ int tty_lines;
 int tty_columns;
 char *tty_device;
 
+static unsigned char *tty_key_seq;
+
 static int tty_device_length;
 static int tty_last_char_flag;
 static int tty_cursor_x;
 static int tty_cursor_y;
 
 /*
- * tty_*_current_attribute:	bit 7:		reverse video
- *				bit 6:		brightness
- *				bits 5,4,3:	background color
- *				bits 2,1,0:	foreground color
+ * tty_*_current_attribute:
+ *   bit 7:		reverse video
+ *   bit 6:		brightness
+ *   bits 5,4,3:	background color
+ *   bits 2,1,0:	foreground color
  */
 static unsigned char tty_current_attribute;
 static unsigned char tty_io_current_attribute;
@@ -237,8 +240,8 @@ tty_key_t default_key;
 
 /* 1Kb of cache memory for terminal optimizations.  Don't make this
    bigger, some terminal/OS combinations fail to transmit chunks of
-   data that are too big.  I'm not sure I fully understand the problem
-   here, but 10Kb didn't work and 1Kb did.  xwrite() reported success,
+   data that are too big.  I'm not sure I fully understand the problem,
+   but 10Kb didn't work and 1Kb did.  xwrite() reported success,
    but the screen was not correctly updated.  xterm bug!?  */
 #define TTY_CACHE_SIZE 1024
 
@@ -517,7 +520,9 @@ unsigned char key_ctrl_table[0x5f] =
 static int  keyno    = 0;
 static int  keyindex = 0;
 static char keybuf[1024];
-
+static unsigned char keystr[MAX_KEY_LENGTH * 20];
+static int partial = 0;
+static int key_on_display = 0;
 
 /* Hooks that get called when we are going idle and when we stop
    being idle.  */
@@ -537,7 +542,6 @@ tty_set_last_char_flag(last_char_flag)
  * This function is used to switch between canonical and noncanonical
  * terminal modes.
  */
-
 void
 tty_set_mode(mode)
     int mode;
@@ -770,7 +774,6 @@ tty_get_mode()
 /*
  * Set the interrupt character.
  */
-
 void
 tty_set_interrupt_char(c)
     int c;
@@ -807,7 +810,6 @@ tty_set_interrupt_char(c)
 /*
  * Flush the tty cache.
  */
-
 void
 tty_flush()
 {
@@ -820,7 +822,7 @@ tty_flush()
 			   tty_index - bytes_transferred);
 	if (count < 0)
 	    break;
-	    
+
 	bytes_transferred += count;
     }
 
@@ -833,7 +835,6 @@ tty_flush()
  * characters.  Actually we are only storing them in a buffer
  * (tty_cache[]) and flush them later (in tty_flush()).
  */
-
 int
 tty_writec(c)
     int c;
@@ -849,7 +850,6 @@ tty_writec(c)
 /*
  * Send the `cl' sequence to the terminal.
  */
-
 void
 tty_io_clear()
 {
@@ -864,7 +864,6 @@ tty_io_clear()
  * attributes.  If screen is NULL, there was an error, so don't try to
  * restore it.
  */
-
 void
 tty_end(screen)
     char *screen;
@@ -892,18 +891,16 @@ tty_end(screen)
  * Converts a key sequence from the human readable, '^' based form into a
  * machine usable form.
  */
-
 char *
-tty_key_convert(key_seq)
+tty_key_human2machine(key_seq)
     unsigned char *key_seq;
 {
     unsigned char *first;
     unsigned char *second;
 
-
     first = second = key_seq;
 
-    if (tty_kbdmode == 0 && *key_seq != '^')
+    if (tty_kbdmode == TTY_RESTRICTED_INPUT && *key_seq != '^')
 	return NULL;
 
     while (*second)
@@ -931,10 +928,51 @@ tty_key_convert(key_seq)
 }
 
 
+/*
+ * Converts a partial key sequence from the machine form into the human
+ * readable, '^' based form.
+ */
+unsigned char *
+tty_key_machine2human(key_seq)
+    char *key_seq;
+{
+    unsigned char *ptr;
+
+    keystr[0] = '\0';
+
+    for (ptr = (unsigned char *)key_seq; *ptr; ptr++)
+    {
+	if (ptr != (unsigned char *)key_seq)
+	    strcat((char *)keystr, " ");
+
+	if (*ptr == key_ESC)
+	    strcat((char *)keystr, "escape");
+	else if (*ptr == ' ')
+	    strcat((char *)keystr, "space");
+	else if (iscntrl(*ptr))
+	{
+	    char x[3];
+	    x[0] = '^';
+	    x[1] = *ptr + 'A' - 1;
+	    x[2] = '\0';
+	    strcat((char *)keystr, x);
+	}
+	else
+	{
+	    char x[2];
+	    x[0] = *ptr;
+	    x[1] = '\0';
+	    strcat((char *)keystr, x);
+	}
+    }
+
+    return (unsigned char *)keystr;
+}
+
+
 /*
  * Update the tty screen.
  */
-
 void
 tty_update()
 {
@@ -963,7 +1001,7 @@ tty_update()
 		tty_io_goto(tty_io_cursor_y = y, tty_io_cursor_x = x);
 
 	    /* Output the color sequence, if necessary, then the
-               character.  */
+	       character.  */
 	    tty_io_colors(tty_atr[pos]);
 	    tty_writec(tty_scr[pos]);
 
@@ -996,7 +1034,6 @@ tty_update()
  * be no attempt to write more than TTY_CACHE_SIZE character in a single
  * call.
  */
-
 static int
 tty_writes(s, len)
     char *s;
@@ -1017,7 +1054,6 @@ tty_writes(s, len)
  * position and the end of the line, it is truncated (i.e. it doesn't
  * wrap arround).  Return the number of characters written.
  */
-
 int
 tty_puts(buf, length)
     char *buf;
@@ -1050,7 +1086,6 @@ tty_puts(buf, length)
 /*
  * Write a character to the screen.
  */
-
 int
 tty_putc(c)
     int c;
@@ -1063,7 +1098,6 @@ tty_putc(c)
 /*
  * Read data from the terminal.
  */
-
 int
 tty_read(buf, length)
     char *buf;
@@ -1093,7 +1127,6 @@ tty_read(buf, length)
  * this feature under Linux, to avoid potential problems, because I
  * don't know which approach is correct.
  */
-
 void
 tty_clear()
 {
@@ -1112,7 +1145,6 @@ tty_clear()
 /*
  * Fill the terminal screen with spaces & the current attribute.
  */
-
 void
 tty_fill()
 {
@@ -1130,7 +1162,6 @@ tty_fill()
  * current content of the screen will be completely different at
  * update time so that the entire screen will be updated.
  */
-
 void
 tty_touch()
 {
@@ -1141,7 +1172,6 @@ tty_touch()
 /*
  * Move the cursor.
  */
-
 static void
 tty_io_goto(y, x)
     int y, x;
@@ -1165,7 +1195,6 @@ tty_io_goto(y, x)
  * Set the foreground color. Use the ANSI color sequence where possible or
  * tty_reverse() for monochrome terminals.
  */
-
 static void
 tty_io_foreground(color)
     int color;
@@ -1194,7 +1223,6 @@ tty_io_foreground(color)
  * Set the background color. Use the ANSI color sequence where possible or
  * tty_reverse() for monochrome terminals.
  */
-
 static void
 tty_io_background(color)
     int color;
@@ -1222,7 +1250,6 @@ tty_io_background(color)
 /*
  * Set the brightness status. See the comment below.
  */
-
 static void
 tty_io_brightness(status)
     int status;
@@ -1272,7 +1299,6 @@ tty_io_brightness(status)
  * Set the reverse video status. This is only used internally by the
  * code in this file therefore it is declared 'static'.
  */
-
 static void
 tty_io_reversevid(status)
     int status;
@@ -1312,7 +1338,6 @@ tty_io_reversevid(status)
 /*
  * Set the brightness, foreground and background altogether.
  */
-
 static void
 tty_io_colors(attributes)
     int attributes;
@@ -1326,7 +1351,6 @@ tty_io_colors(attributes)
 /*
  * Move the cursor.
  */
-
 void
 tty_goto(y, x)
     int y, x;
@@ -1339,7 +1363,6 @@ tty_goto(y, x)
 /*
  * Return the current coordinates of the cursor (line/column).
  */
-
 void
 tty_get_cursor(y, x)
     int *y, *x;
@@ -1352,7 +1375,6 @@ tty_get_cursor(y, x)
 /*
  * Set the foreground color.
  */
-
 void
 tty_foreground(color)
     int color;
@@ -1364,7 +1386,6 @@ tty_foreground(color)
 /*
  * Set the background color.
  */
-
 void
 tty_background(color)
     int color;
@@ -1376,7 +1397,6 @@ tty_background(color)
 /*
  * Set the brightness status. See the comment below.
  */
-
 void
 tty_brightness(status)
     int status;
@@ -1389,7 +1409,6 @@ tty_brightness(status)
  * Set the reverse video status. This is only used internally by the
  * code in this file therefore it is declared 'static'.
  */
-
 void
 tty_reversevid(status)
     int status;
@@ -1401,7 +1420,6 @@ tty_reversevid(status)
 /*
  * Set the brightness, foreground and background all together.
  */
-
 void
 tty_colors(brightness, foreground, background)
     int brightness, foreground, background;
@@ -1415,7 +1433,6 @@ tty_colors(brightness, foreground, background)
 /*
  * Beep :-)
  */
-
 void
 tty_beep()
 {
@@ -1428,7 +1445,6 @@ tty_beep()
  * Set the cursor status (where possible). Make it invisible during screen
  * refreshes and restore it afterward.
  */
-
 void
 tty_cursor(status)
     int status;
@@ -1449,7 +1465,6 @@ tty_cursor(status)
 /*
  * Store the software terminal status in a tty_status_t structure.
  */
-
 void
 tty_save(status)
     tty_status_t *status;
@@ -1461,7 +1476,6 @@ tty_save(status)
 /*
  * Restore the software terminal status from a tty_status_t structure.
  */
-
 void
 tty_restore(status)
     tty_status_t *status;
@@ -1473,7 +1487,6 @@ tty_restore(status)
 /*
  * Restore the terminal to its default state.
  */
-
 void
 tty_defaults()
 {
@@ -1497,7 +1510,6 @@ tty_defaults()
  * FIXME: It seems that the 8th bit can be used for parity as well.
  * This case is not handled yet.  It should be configurable.
  */
-
 static int
 tty_extract_key()
 {
@@ -1519,7 +1531,6 @@ tty_extract_key()
  * Get a character from the terminal.  For better performance on
  * highly loaded systems, read all the data available.
  */
-
 int
 tty_getc()
 {
@@ -1534,7 +1545,8 @@ tty_getc()
     /* Allow signals to suspend/resize git.  */
     signals(ON);
 
-    for (keyindex = 0; (keyno = tty_read(keybuf, 1024)) < 0;)
+    keyindex = 0;
+    while ((keyno = tty_read(keybuf, 1024)) < 0)
 	;
 
     /* Prevent signals from suspending/resizing git.  */
@@ -1550,7 +1562,6 @@ tty_getc()
 /*
  * Insert a key sequence into the list.
  */
-
 void
 tty_key_list_insert_sequence(key, key_seq, aux_data)
     tty_key_t **key;
@@ -1571,7 +1582,6 @@ tty_key_list_insert_sequence(key, key_seq, aux_data)
  * Parse the key list, inserting the new key sequence in the proper
  * position.
  */
-
 void
 tty_key_list_insert(key_seq, aux_data)
     unsigned char *key_seq;
@@ -1600,10 +1610,20 @@ tty_key_list_insert(key_seq, aux_data)
 
 
 /*
- * Incremental searches a key in the list. Returns a pointer to the first
- * sequence that matches.
+ * Force the next key search to start from the begining of the list.
  */
+void
+tty_key_search_restart()
+{
+    current_key = key_list_head;
+}
 
+
+/*
+ * Incremental searches a key in the list. Returns a pointer to the first
+ * sequence that matches, -1 if there is no match and NULL if there can
+ * still be a match.
+ */
 tty_key_t *
 tty_key_search(key_seq)
     char *key_seq;
@@ -1625,28 +1645,18 @@ tty_key_search(key_seq)
     }
 
     if (current_key == NULL ||
-	strncmp(key_seq, (char *)current_key->key_seq, strlen(key_seq)))
+	strncmp(key_seq, (char *)current_key->key_seq, strlen(key_seq)) != 0)
 	return (tty_key_t *)-1;
     else
 	return NULL;
 }
 
 
-/*
- * Force the next key search to start from the begining of the list.
- */
-
-void
-tty_key_search_restart()
-{
-    current_key = key_list_head;
-}
-
-
 #if 0
-/* Delete a key from the list. Don't remove this function, God only knows
-   when I'll gonna need it...  */
-
+/*
+ * Delete a key from the list. Don't remove this function, God only knows
+ * when I'm gonna need it...
+ */
 void
 tty_key_list_delete()
 {
@@ -1662,26 +1672,61 @@ tty_key_list_delete()
 #endif
 
 
+/*
+ * Print the key sequence on the last line of the screen.
+ */
+void
+tty_key_print(key_seq)
+    char *key_seq;
+{
+    tty_status_t tty_status;
+    char typed[] = "Keys typed so far: ";
+    char incomplete[] = " ";
+    char *spaces;
+
+    tty_save(&tty_status);
+    tty_goto(tty_lines - 1, 0);
+    tty_background(WHITE);
+    tty_foreground(BLACK);
+
+    spaces = xmalloc(tty_columns+1);
+    memset(spaces, ' ', tty_columns);
+    spaces[tty_columns] = '\0';
+    tty_puts(spaces, tty_columns);
+    xfree(spaces);
+    tty_goto(tty_lines - 1, 0);
+
+    tty_key_machine2human(key_seq);
+
+    tty_puts(typed, strlen(typed));
+    tty_puts((char *)keystr, strlen((char *)keystr));
+    tty_puts(incomplete, strlen(incomplete));
+
+    tty_update();
+    tty_restore(&tty_status);
+}
+
+
 /*
  * Get the first key available, returning also the repeat count, that
  * is, the number of times the key has been pressed.  These feature is
  * mainly used by the calling routines to perform optimizations.  For
  * example, if you press the down arrow several times, the caller can
  * display only the final position, saving a lot of time. If you have
- * ever worked with git on high loaded systems, I'm sure you know what
+ * ever worked with git on highly loaded systems, I'm sure you know what
  * I mean.
+ *
+ * If tty_get_key() returns NULL, the key sequence is invalid.
  */
-
 tty_key_t *
 tty_get_key(repeat_count)
     int *repeat_count;
 {
     int i, c;
     tty_key_t *key;
-    unsigned char key_seq[64];
 
-  restart:
-    while ((c = tty_getc()) == -1);
+    while ((c = tty_getc()) == -1)
+	;
 
     if (repeat_count)
 	*repeat_count = 1;
@@ -1690,7 +1735,7 @@ tty_get_key(repeat_count)
     if (c == 0)
 	c = 0xff;
 
-    if (!tty_kbdmode)
+    if (tty_kbdmode == TTY_RESTRICTED_INPUT)
     {
 	if (c == '\n' || c == '\r')
 	    c = key_ENTER;
@@ -1698,51 +1743,105 @@ tty_get_key(repeat_count)
 	if (isprint(c) || c == key_INTERRUPT)
 	{
 	    default_key.key_seq[0] = c;
-	    default_key.key_seq[1] = 0;
+	    default_key.key_seq[1] = '\0';
 	    return &default_key;
 	}
     }
 
+    partial = 0;
+    key_on_display = 0;
+
     tty_key_search_restart();
 
     for (i = 0; i < MAX_KEY_LENGTH; i++)
     {
-	/* Handle brain-damaged key sequences correctly.  If a 0 is
-	   found, transform it in a 0xff.  I don't know how smart this
-	   is, but right know I don't feel like doing it otherwise.  */
+	/* Kludge to handle brain-damaged key sequences.  If a 0
+	   (^SPC) is found, change it into 0xff.  I don't know how
+	   smart this is, but right know I don't feel like doing it
+	   otherwise.  */
 	if (c == 0)
 	    c = 0xff;
 
-	key_seq[i    ] = c;
-	key_seq[i + 1] = 0;
+	tty_key_seq[i] = c;
+	tty_key_seq[i + 1] = 0;
 
-	key = tty_key_search((char *)key_seq);
+	key = tty_key_search((char *)tty_key_seq);
 
 	if (key == (tty_key_t *)-1)
-	    goto restart;
+	{
+	    alarm(1);
+	    partial = 0;
+	    return NULL;
+	}
 
 	if (key)
+	{
+	    if (partial)
+	    {
+		/* tty_key_print((char *)tty_key_seq); */
+		/* tty_update(); */
+		/* Small delay for displaying the selected sequence.  */
+		/* FIXME: 1 second is way too much!  */
+		/* sleep(1);  */
+	    }
 	    break;
+	}
+
+	if (keyno == 0)
+	{
+	    /* Schedule an alarm in 1 second, to display the key
+	       sequence if still incomplete by that time.  */
+	    if (key_on_display)
+		tty_key_print((char *)tty_key_seq);
+	    else
+		alarm(1);
+	    partial = 1;
+	}
 
 	while ((c = tty_getc()) == -1)
 	    ;
     }
 
     if (i == MAX_KEY_LENGTH)
-	goto restart;
+    {
+	alarm(1);
+	partial = 0;
+	return NULL;
+    }
 
     if (repeat_count)
-	while (keyno > i && (memcmp(key_seq, &keybuf[keyindex], i + 1) == 0))
+	while (keyno > i &&
+	       (memcmp(tty_key_seq, &keybuf[keyindex], i + 1) == 0))
 	{
 	    keyindex += i + 1;
-	    keyno    -= i + 1;
+	    keyno -= i + 1;
 	    (*repeat_count)++;
 	}
 
+    alarm(1);
+    partial = 0;
     return key;
 }
 
 
+void
+tty_key_print_async()
+{
+    if (partial)
+    {
+	tty_key_print((char *)tty_key_seq);
+	key_on_display = 1;
+    }
+}
+
+
+char *
+tty_get_previous_key_seq()
+{
+    return (char *)tty_key_seq;
+}
+
+
 #define columns_ok(x) (((x) > 0) && ((x) <= MAX_TTY_COLUMNS))
 #define lines_ok(x)   (((x) > 0) && ((x) <= MAX_TTY_LINES))
 
@@ -1757,7 +1856,6 @@ tty_get_key(repeat_count)
  * Note that we try to get the best value separately for lines and for
  * columns.
  */
-
 void
 tty_resize()
 {
@@ -1872,7 +1970,6 @@ tty_resize()
  * easier (using /de/vcsa0 when debugging dumps the screen on the
  * debugger's console; not funny).
  */
-
 int
 vcs_io(buf, op)
     char *buf;
@@ -1970,7 +2067,6 @@ vcs_io(buf, op)
  * Dump the screen. Only used in Linux if the terminal is a virtual
  * console.
  */
-
 void
 tty_get_screen(buf)
     char *buf;
@@ -1978,6 +2074,8 @@ tty_get_screen(buf)
 #ifdef HAVE_LINUX
     if (LinuxConsole)
 	vcs_read_ok = vcs_io(buf, VCS_READ);
+#else
+    buf = NULL;
 #endif  /* HAVE_LINUX */
 }
 
@@ -2017,6 +2115,7 @@ tty_put_screen(buf)
 	tty_clear();
 #else   /* !HAVE_LINUX */
     tty_clear();
+    buf = NULL;
 #endif  /* !HAVE_LINUX */
 }
 
@@ -2024,7 +2123,6 @@ tty_put_screen(buf)
 /*
  * Get the color index from the colors[] index table.
  */
-
 int
 tty_get_color_index(colorname)
     char *colorname;
@@ -2043,7 +2141,6 @@ tty_get_color_index(colorname)
  * Get the capability of a given termcap symbol.  Return NULL if there
  * is no capability for it.
  */
-
 char *
 tty_get_symbol_key_seq(symbol)
     char *symbol;
@@ -2059,10 +2156,9 @@ tty_get_symbol_key_seq(symbol)
 
 
 /*
- * Get the entire set of requiresd termcap/terminfo capabilities. It performs
+ * Get the entire set of required termcap/terminfo capabilities. It performs
  * consistency checkings trying to recover as well as possible.
  */
-
 void
 tty_get_capabilities()
 {
@@ -2122,10 +2218,12 @@ tty_get_capabilities()
 
 	/* Try something resonable, given the current value.  Add here
 	   as needed...  */
-	if (strcmp(termtype, "xterm-color") == 0 ||
-	    strcmp(termtype, "xterms")      == 0 ||
-	    strcmp(termtype, "aixterm")     == 0 ||
-	    strcmp(termtype, "dtterm")      == 0)
+	if (strcmp(termtype, "xterm-color")  == 0 ||
+	    strcmp(termtype, "xterms")       == 0 ||
+	    strcmp(termtype, "xterm-debian") == 0 ||
+	    strcmp(termtype, "rxvt")         == 0 ||
+	    strcmp(termtype, "aixterm")      == 0 ||
+	    strcmp(termtype, "dtterm")       == 0)
 	{
 	    fprintf(stderr, "%s: trying xterm...\n", program);
 	    termtype = "xterm";
@@ -2241,7 +2339,6 @@ tty_get_capabilities()
 /*
  * Initialize the tty driver.
  */
-
 void
 tty_init(kbd_mode)
     int kbd_mode;
@@ -2273,10 +2370,10 @@ tty_init(kbd_mode)
 #endif /* HAVE_SYSTEMV_TTY */
 #endif /* HAVE_POSIX_TTY */
 
-    default_key.key_seq  = (unsigned char *)xmalloc(16);
+    default_key.key_seq  = tty_key_seq = (unsigned char *)xmalloc(64);
     default_key.aux_data = NULL;
-    default_key.next     = NULL;
-    tty_kbdmode          = kbd_mode;
+    default_key.next = NULL;
+    tty_kbdmode = kbd_mode;
 
     tty_device_length = strlen(tty_device);
     tty_get_capabilities();
@@ -2286,7 +2383,6 @@ tty_init(kbd_mode)
 /*
  * Update the title of xterm-like X terminal emulators.
  */
-
 void
 tty_update_title(string)
     char *string;
